from .codeblock import CodeBlockParser, PythonCodeBlockParser
from .doctest import DocTestDirectiveParser
from .skip import SkipParser
from .clear import ClearNamespaceParser

__all__ = [
    'CodeBlockParser',
    'PythonCodeBlockParser',
    'DocTestDirectiveParser',
    'SkipParser',
    'ClearNamespaceParser',
]



from typing import Any, Optional

from sybil.typing import Evaluator, LexemeMapping


class Lexeme(str):
    """
    Where needed, this can store both the text of the lexeme
    and it's line offset relative to the line number of the example
    that contains it.
    """

    def __new__(cls, text: str, offset: int, line_offset: int) -> 'Lexeme':
        return str.__new__(cls, text)

    def __init__(self, text: str, offset: int, line_offset: int) -> None:
        self.text = text
        self.offset = offset
        self.line_offset = line_offset

    def strip_leading_newlines(self) -> 'Lexeme':
        stripped = self.lstrip('\n')
        removed = len(self) - len(stripped)
        return Lexeme(stripped, self.offset + removed, self.line_offset + removed)


MAX_REPR_PART_LENGTH = 40
CONTRACTED = '...'


class Region:
    """
    Parsers should yield instances of this class for each example they
    discover in a documentation source file.
    
    :param start: 
        The character position at which the example starts in the
        :class:`~sybil.document.Document`.
    
    :param end: 
        The character position at which the example ends in the
        :class:`~sybil.document.Document`.
    
    :param parsed: 
        The parsed version of the example.
    
    :param evaluator: 
        The callable to use to evaluate this example and check if it is
        as it should be.
    """

    def __init__(
            self,
            start: int,
            end: int,
            parsed: Any = None,
            evaluator: Optional[Evaluator] = None,
            lexemes: Optional[LexemeMapping] = None,
    ) -> None:
        #: The start of this region within the document's :attr:`~sybil.Document.text`.
        self.start: int = start
        #: The end of this region within the document's :attr:`~sybil.Document.text`.
        self.end: int = end
        #: The parsed version of this region. This only needs to have meaning to
        #: the :attr:`evaluator`.
        self.parsed: Any = parsed
        #: The :any:`Evaluator` for this region.
        self.evaluator: Optional[Evaluator] = evaluator
        #: The lexemes extracted from the region.
        self.lexemes: LexemeMapping = lexemes or {}

    @staticmethod
    def trim(text: str) -> str:
        if len(text) > MAX_REPR_PART_LENGTH:
            half = int((MAX_REPR_PART_LENGTH + len(CONTRACTED)) / 2)
            text = text[:half] + CONTRACTED + text[-half:]
        return text

    def __repr__(self) -> str:
        evaluator_text = f' evaluator={self.evaluator!r}' if self.evaluator else ''
        text = f'<Region start={self.start} end={self.end}{evaluator_text}>'
        if self.lexemes:
            text += '\n'
        for name, lexeme in self.lexemes.items():
            if isinstance(lexeme, str):
                lexeme = self.trim(lexeme)
            text += f'{name}: {lexeme!r}\n'
        if self.parsed:
            parsed_text = self.trim(repr(self.parsed))
            text += f'<Parsed>{parsed_text}</Parsed>'
        if self.parsed or self.lexemes:
            text += '</Region>'
        return text

    def __lt__(self, other: 'Region') -> bool:
        assert isinstance(other, type(self)), f"{type(other)} not supported for <"
        assert self.start == other.start  # This is where this may happen, if not something weird
        return True

    def adjust(self, lexed: 'Region', lexeme: Lexeme) -> None:
        """
        Adjust the start and end of this region based on the provided :class:`Lexeme`
        and ::class:`Region` that lexeme came from.
        """
        self.start += (lexed.start + lexeme.offset)
        self.end += lexed.start



import sys
from importlib import import_module
from pathlib import Path

import pytest

from sybil.python import import_cleanup
from .helpers import Finder, ast_docstrings


def test_finder_present_but_is_not_present():
    finder = Finder('foo\nbaz\n')
    with pytest.raises(AssertionError) as info:
        finder.assert_present('bob')
    assert str(info.value) == "foo\nbaz\n\n'foo\\nbaz\\n'"


def test_finder_not_present_but_is_present():
    finder = Finder('foo baz bar')
    with pytest.raises(AssertionError) as info:
        finder.assert_not_present('baz')
    assert str(info.value) == '\nfoo baz bar'


def test_import_cleanup(tmp_path: Path):
    (tmp_path / 'some_module.py').write_text('import unittest\nfoo = 1')
    (tmp_path / 'other_module.py').write_text('import other_module\nfoo = 1')

    initial_modules = sys.modules.copy()
    initial_path = sys.path.copy()

    with import_cleanup():
        sys.path.append(str(tmp_path))
        some_module = import_module('some_module')
        assert some_module.foo == 1

    assert sys.modules == initial_modules
    assert sys.path == initial_path


def test_all_python_files(all_python_files):
    count = len(all_python_files)
    assert count > 50, count


def test_ast_docstrings(all_python_files):
    seen_docstrings = 0
    for _, source in all_python_files:
        seen_docstrings += len(tuple(ast_docstrings(source)))
    assert seen_docstrings > 50, seen_docstrings



# coding=utf-8
from doctest import REPORT_NDIFF, ELLIPSIS

import pytest
from testfixtures import compare

from sybil.example import SybilFailure
from sybil.parsers.rest import DocTestParser as ReSTDocTestParser
from sybil.parsers.myst import DocTestDirectiveParser, PythonCodeBlockParser
from tests.helpers import parse, sample_path


def test_use_existing_doctest_parser():
    path = sample_path('myst-doctest-use-rest.md')
    examples, namespace = parse('myst-doctest-use-rest.md', ReSTDocTestParser(), expected=6)
    examples[0].evaluate()
    assert namespace['x'] == 2
    examples[1].evaluate()
    assert namespace['x'] == 2
    examples[2].evaluate()
    assert namespace['x'] == 3
    with pytest.raises(SybilFailure) as excinfo:
        examples[3].evaluate()
    compare(str(excinfo.value), expected = (
        f"Example at {path}, line 28, column 1 did not evaluate as expected:\n"
        "Expected:\n"
        "    3\n"
        "Got:\n"
        "    2\n"
    ))
    examples[4].evaluate()
    examples[5].evaluate()


def test_use_python_codeblock_parser():
    examples, namespace = parse('myst-doctest.md', PythonCodeBlockParser(), expected=3)
    examples[0].evaluate()
    examples[1].evaluate()
    assert namespace['x'] == 2
    examples[2].evaluate()
    assert namespace['x'] == 3


def test_fail_with_options_using_python_codeblock_parser():
    parser = PythonCodeBlockParser(doctest_optionflags=REPORT_NDIFF|ELLIPSIS)
    examples, namespace = parse('myst-doctest-fail.md', parser, expected=1)
    with pytest.raises(SybilFailure) as excinfo:
        examples[0].evaluate()
    assert excinfo.value.result == (
        "Differences (ndiff with -expected +actual):\n"
        "    - Not my output\n"
        "    + where's my output?\n"
    )


def test_use_doctest_role_parser():
    path = sample_path('myst-doctest.md')
    examples, namespace = parse('myst-doctest.md', DocTestDirectiveParser(), expected=4)
    examples[0].evaluate()
    assert namespace['y'] == 2
    examples[1].evaluate()
    examples[2].evaluate()
    with pytest.raises(SybilFailure) as excinfo:
        examples[3].evaluate()
    compare(str(excinfo.value), expected=(
        f"Example at {path}, line 45, column 1 did not evaluate as expected:\n"
        "Expected:\n"
        "    3\n"
        "Got:\n"
        "    2\n"
    ))


def test_fail_with_options_using_doctest_role_parser():
    parser = DocTestDirectiveParser(optionflags=REPORT_NDIFF|ELLIPSIS)
    examples, namespace = parse('myst-doctest-fail.md', parser, expected=1)
    with pytest.raises(SybilFailure) as excinfo:
        examples[0].evaluate()
    assert excinfo.value.result == (
        "Differences (ndiff with -expected +actual):\n"
        "    - Also not my output\n"
        "    + where's my output?\n"
    )



from sybil.parsers.myst import ClearNamespaceParser, PythonCodeBlockParser
from .helpers import parse


def test_basic():
    examples, namespace = parse(
        'myst-clear.md', PythonCodeBlockParser(), ClearNamespaceParser(), expected=4
    )
    for example in examples:
        example.evaluate()
    assert 'x' not in namespace, namespace



# coding=utf-8
from doctest import REPORT_NDIFF, ELLIPSIS, DocTestParser as BaseDocTestParser
from pathlib import Path

import pytest
from testfixtures import compare

from sybil.document import Document, PythonDocStringDocument
from sybil.example import SybilFailure
from sybil.parsers.abstract import DocTestStringParser
from sybil.parsers.rest import DocTestParser, DocTestDirectiveParser
from .helpers import sample_path, parse, FUNCTIONAL_TEST_DIR


def test_pass():
    examples, namespace = parse('doctest.txt', DocTestParser(), expected=5)
    examples[0].evaluate()
    assert namespace['y'] == 1
    examples[1].evaluate()
    assert namespace['y'] == 1
    examples[2].evaluate()
    assert namespace['x'] == [1, 2, 3]
    examples[3].evaluate()
    assert namespace['y'] == 2
    examples[4].evaluate()
    assert namespace['y'] == 2


def test_fail():
    path = sample_path('doctest_fail.txt')
    examples, namespace = parse('doctest_fail.txt', DocTestParser(), expected=2)
    with pytest.raises(SybilFailure) as excinfo:
        examples[0].evaluate()
    # Note on line numbers: This test shows how the example's line number is correct
    # however, doctest doesn't do the line padding trick Sybil does with codeblocks,
    # so the line number will never be correct, it's always 1.
    compare(str(excinfo.value), expected=(
        f"Example at {path}, line 1, column 1 did not evaluate as expected:\n"
        "Expected:\n"
        "    Not my output\n"
        "Got:\n"
        "    where's my output?\n"
    ))
    with pytest.raises(SybilFailure) as excinfo:
        examples[1].evaluate()
    actual = excinfo.value.result
    assert actual.startswith('Exception raised:')
    assert actual.endswith('Exception: boom!\n')


def test_fail_with_options():
    parser = DocTestParser(optionflags=REPORT_NDIFF|ELLIPSIS)
    examples, namespace = parse('doctest_fail.txt', parser, expected=2)
    with pytest.raises(SybilFailure) as excinfo:
        examples[0].evaluate()
    assert excinfo.value.result == (
        "Differences (ndiff with -expected +actual):\n"
        "    - Not my output\n"
        "    + where's my output?\n"
    )


def test_literals():
    parser = DocTestParser()
    examples, _ = parse('doctest_literals.txt', parser, expected=5)
    for example in examples:
        example.evaluate()


def test_min_indent():
    examples, _ = parse('doctest_min_indent.txt', DocTestParser(), expected=1)
    examples[0].evaluate()


def test_tabs():
    path = sample_path('doctest_tabs.txt')
    parser = DocTestParser()
    with pytest.raises(ValueError):
        Document.parse(path, parser)


def test_irrelevant_tabs():
    examples, _ = parse('doctest_irrelevant_tabs.txt', DocTestParser(), expected=1)
    examples[0].evaluate()


def test_unicode():
    examples, _ = parse('doctest_unicode.txt', DocTestParser(), expected=1)
    examples[0].evaluate()


def test_directive():
    path = sample_path('doctest_directive.txt')
    examples, _ = parse('doctest_directive.txt', DocTestDirectiveParser(), expected=3)
    examples[0].evaluate()
    with pytest.raises(SybilFailure) as excinfo:
        examples[1].evaluate()
    compare(str(excinfo.value), expected = (
        f"Example at {path}, line 13, column 1 did not evaluate as expected:\n"
        "Expected:\n"
        "    Unexpected!\n"
        "Got:\n"
        "    2\n"
    ))
    with pytest.raises(SybilFailure) as excinfo:
        examples[2].evaluate()
    actual = excinfo.value.result
    assert actual.startswith('Exception raised:')
    assert actual.endswith('Exception: boom!\n')


def test_directive_with_options():
    path = sample_path('doctest_directive.txt')
    parser = DocTestDirectiveParser(optionflags=REPORT_NDIFF|ELLIPSIS)
    examples, namespace = parse('doctest_directive.txt', parser, expected=3)
    with pytest.raises(SybilFailure) as excinfo:
        examples[1].evaluate()
    compare(str(excinfo.value), expected = (
        f"Example at {path}, line 13, column 1 did not evaluate as expected:\n"
        "Differences (ndiff with -expected +actual):\n"
        "    - Unexpected!\n"
        "    + 2\n"
    ))


# Number of doctests that can't be parsed in a file when looking at the whole file source:
ROOT = Path(FUNCTIONAL_TEST_DIR)
UNPARSEABLE = {
   ROOT / 'package_and_docs' / 'src' / 'parent' / 'child' / 'module_b.py': 1,
}
MINIMUM_EXPECTED_DOCTESTS = 9


def test_sybil_example_count(all_python_files):
    parser = DocTestStringParser()

    seen_examples_from_source = 0
    seen_examples_from_docstrings = 0

    for path, source in all_python_files:
        seen_examples_from_source += len(tuple(parser(source, path)))
        for start, end, docstring in PythonDocStringDocument.extract_docstrings(source):
            seen_examples_from_docstrings += len(tuple(parser(docstring, path)))

    assert seen_examples_from_source > MINIMUM_EXPECTED_DOCTESTS, seen_examples_from_source
    assert seen_examples_from_docstrings > MINIMUM_EXPECTED_DOCTESTS, seen_examples_from_docstrings
    assert seen_examples_from_docstrings == seen_examples_from_source


def check_sybil_against_doctest(path, text):
    problems = []
    name = str(path)
    regions = list(DocTestStringParser()(text, path))
    sybil_examples = [region.parsed for region in regions]
    doctest_examples = BaseDocTestParser().get_examples(text, name)
    problems.append(compare(
        expected=doctest_examples, actual=sybil_examples, raises=False, show_whitespace=True
    ))
    for region in regions:
        example = region.parsed
        region_source = text[region.start:region.end]
        for name in 'source', 'want':
            expected = getattr(example, name)
            if name == 'source':
                # doctest.DocTestParser REPL examples are pre-processed into python
                # source to execute, so do similar here but note this is pretty fragile:
                expected_source = region_source.replace('>>>', '').replace('    ... ', '')
            else:
                expected_source = region_source
            if expected not in expected_source:  # pragma: no cover - Only on failures!
                problems.append(f'{region}:{name}\n{expected!r} not in {expected_source!r}')
    problems = [problem for problem in problems if problem]
    if problems:  # pragma: no cover - Only on failures!
        raise AssertionError('doctests not correctly extracted:\n\n'+'\n\n'.join(problems))


def test_all_docstest_examples_extracted_from_source_correctly(python_file):
    path, source = python_file
    if path in UNPARSEABLE:
        return
    check_sybil_against_doctest(path, source)


def test_all_docstest_examples_extracted_from_docstrings_correctly(python_file):
    path, source = python_file
    for start, end, docstring in PythonDocStringDocument.extract_docstrings(source):
        check_sybil_against_doctest(path, docstring)



import sys
from pathlib import Path

import pytest
from pytest import CaptureFixture
from testfixtures import compare

from sybil import Sybil
from sybil.parsers.rest import PythonCodeBlockParser, DocTestParser
from sybil.python import import_cleanup
from .helpers import (
    run_pytest, run_unittest, PYTEST, run, write_config, UNITTEST, write_doctest,
    functional_sample, clone_functional_sample, check_path, sample_path
)


@pytest.fixture(autouse=True)
def cleanup_imports():
    with import_cleanup():
        yield


def test_pytest(capsys: CaptureFixture[str]):
    results = run_pytest(capsys, functional_sample('pytest'))
    out = results.out

    # check we're trimming tracebacks:
    out.assert_not_present('sybil/example.py')

    out.then_find('fail.rst::line:1,column:1')
    out.then_find('fail.rst::line:1,column:1 sybil setup session_fixture setup\n'
                  'module_fixture setup\n'
                  'class_fixture setup\n'
                  'function_fixture setup\n'
                  'x is currently: 0\n'
                  'FAILED function_fixture teardown\n'
                  'class_fixture teardown')
    out.then_find('fail.rst::line:6,column:1')
    out.then_find('fail.rst::line:6,column:1 class_fixture setup\n'
                  'function_fixture setup\n'
                  '0smcf PASSED function_fixture teardown\n'
                  'class_fixture teardown')
    out.then_find('fail.rst::line:8,column:1')
    out.then_find('fail.rst::line:8,column:1 class_fixture setup\n'
                  'function_fixture setup\n'
                  '1smcf FAILED function_fixture teardown\n'
                  'class_fixture teardown')
    out.then_find('fail.rst::line:10,column:1')
    out.then_find('fail.rst::line:10,column:1 class_fixture setup\n'
                  'function_fixture setup\n'
                  '2smcf FAILED function_fixture teardown\n'
                  'class_fixture teardown')
    out.then_find('fail.rst::line:12,column:1')
    out.then_find('fail.rst::line:12,column:1 class_fixture setup\n'
                  'function_fixture setup\n'
                  '3smcf PASSED function_fixture teardown\n'
                  'class_fixture teardown')
    out.then_find('fail.rst::line:14,column:1')
    out.then_find('fail.rst::line:14,column:1 class_fixture setup\n'
                  'function_fixture setup\n'
                  'FAILED function_fixture teardown\n'
                  'class_fixture teardown\n'
                  'module_fixture teardown\n'
                  'sybil teardown 5')
    out.then_find('pass.rst::line:1,column:1')
    out.then_find('pass.rst::line:1,column:1 sybil setup module_fixture setup\n'
                  'class_fixture setup\n'
                  'function_fixture setup\n'
                  '0smcf PASSED function_fixture teardown\n'
                  'class_fixture teardown')
    out.then_find('pass.rst::line:3,column:1')
    out.then_find('pass.rst::line:3,column:1 class_fixture setup\n'
                  'function_fixture setup\n'
                  '1smcf PASSED function_fixture teardown\n'
                  'class_fixture teardown')
    out.then_find('pass.rst::line:5,column:1')
    out.then_find('pass.rst::line:5,column:1 class_fixture setup\n'
                  'function_fixture setup\n'
                  '2smcf PASSED function_fixture teardown\n'
                  'class_fixture teardown')
    out.then_find('pass.rst::line:7,column:1')
    out.then_find('pass.rst::line:7,column:1 class_fixture setup\n'
                  'function_fixture setup\n'
                  '3smcf PASSED function_fixture teardown\n'
                  'class_fixture teardown\n'
                  'module_fixture teardown\n'
                  'sybil teardown 4\n'
                  'session_fixture teardown')
    out.then_find('_ fail.rst line=1 column=1 _')
    out.then_find("raise Exception('the start!')")
    out.then_find('_ fail.rst line=8 column=1 _')
    out.then_find('Y count was 3 instead of 2')
    out.then_find('fail.rst:8: SybilFailure')
    out.then_find('_ fail.rst line=10 column=1 _')
    out.then_find('ValueError: X count was 3 instead of 4')
    out.then_find('_ fail.rst line=14 column=1 _')
    out.then_find("raise Exception('boom!')")
    out.then_find('fail.rst:18: Exception')

    assert results.return_code == 1
    assert results.failures == 4
    assert results.total == 10


def test_unittest(capsys: CaptureFixture[str]):
    results = run_unittest(capsys, functional_sample('unittest'))
    out = results.out
    out.then_find('sybil setup')
    out.then_find('fail.rst,line:6,column:1 ... 0\nok')
    out.then_find('fail.rst,line:8,column:1 ... 1\nFAIL')
    out.then_find('fail.rst,line:10,column:1 ... 2\nERROR')
    out.then_find('fail.rst,line:12,column:1 ... 3\nok')
    out.then_find('sybil teardown 4\nsybil setup')
    out.then_find('pass.rst,line:1,column:1 ... 0\nok')
    out.then_find('pass.rst,line:3,column:1 ... 1\nok')
    out.then_find('pass.rst,line:5,column:1 ... 2\nok')
    out.then_find('pass.rst,line:7,column:1 ... 3\nok')
    out.then_find('sybil teardown 4')
    out.then_find('ERROR: ')
    out.then_find('fail.rst,line:10,column:1')
    out.then_find('ValueError: X count was 3 instead of 4')
    out.then_find('FAIL:')
    out.then_find('fail.rst,line:8,column:1')
    out.then_find('Y count was 3 instead of 2')
    out.then_find('Ran 8 tests')

    assert results.total == 8
    assert results.failures == 1
    assert results.errors == 1


def make_tree(tmp_path: Path):
    write_doctest(tmp_path, 'foo.rst')
    write_doctest(tmp_path, 'bar.rst')
    write_doctest(tmp_path, 'parent', 'foo.rst')
    write_doctest(tmp_path, 'parent', 'bar.rst')
    write_doctest(tmp_path, 'parent', 'child', 'foo.rst')
    write_doctest(tmp_path, 'parent', 'child', 'bar.rst')


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_filter_everything(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    make_tree(tmp_path)
    write_config(tmp_path, runner)
    results = run(capsys, runner, tmp_path)
    # ask for nothing, get nothing...
    assert results.total == 0


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_filter_just_pattern(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    make_tree(tmp_path)
    write_config(tmp_path, runner, pattern="'*.rst'")
    results = run(capsys, runner, tmp_path)
    assert results.total == 6


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_filter_fnmatch_pattern(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    make_tree(tmp_path)
    write_config(tmp_path, runner, pattern="'**/*.rst'")
    results = run(capsys, runner, tmp_path)
    # The fact that the two .rst files in the root aren't matched is
    # arguably a bug in the Python interpretation of **/
    results.out.assert_has_run(runner, '/parent/foo.rst')
    results.out.assert_has_run(runner, '/parent/bar.rst')
    results.out.assert_has_run(runner, '/parent/child/foo.rst')
    results.out.assert_has_run(runner, '/parent/child/bar.rst')
    assert results.total == 4, results.out.text


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_filter_just_filenames(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    make_tree(tmp_path)
    write_config(tmp_path, runner,
                 filenames="['bar.rst']")
    results = run(capsys, runner, tmp_path)
    results.out.assert_has_run(runner, '/bar.rst')
    results.out.assert_has_run(runner, '/parent/bar.rst')
    results.out.assert_has_run(runner, '/parent/child/bar.rst')
    assert results.total == 3, results.out.text


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_filter_directory(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    make_tree(tmp_path)
    write_config(tmp_path, runner,
                 path=f"'{tmp_path / 'parent'}'",
                 pattern="'*.rst'")
    results = run(capsys, runner, tmp_path)
    results.out.assert_has_run(runner, '/parent/foo.rst')
    results.out.assert_has_run(runner, '/parent/bar.rst')
    results.out.assert_has_run(runner, '/parent/child/foo.rst')
    results.out.assert_has_run(runner, '/parent/child/bar.rst')
    assert results.total == 4, results.out.text


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_filter_directory_with_excludes(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    make_tree(tmp_path)
    write_config(tmp_path, runner,
                 path=f"'{tmp_path / 'parent'}'",
                 pattern="'*.rst'",
                 exclude="'ba*.rst'")
    results = run(capsys, runner, tmp_path)
    results.out.assert_has_run(runner, '/parent/foo.rst')
    results.out.assert_has_run(runner, '/parent/child/foo.rst')
    assert results.total == 2, results.out.text


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_filter_filenames_and_excludes(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    make_tree(tmp_path)
    write_config(tmp_path, runner,
                 path=f"'{tmp_path / 'parent'}'",
                 filenames="{'bar.rst'}",
                 excludes="['**child/*.rst']")
    results = run(capsys, runner, tmp_path)
    results.out.assert_has_run(runner, '/parent/bar.rst')
    assert results.total == 1, results.out.text


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_filter_exclude_by_name(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    write_doctest(tmp_path, 'foo.txt')
    write_doctest(tmp_path, 'bar.txt')
    write_doctest(tmp_path, 'child', 'foo.txt')
    write_doctest(tmp_path, 'child', 'bar.txt')
    write_config(tmp_path, runner,
                 pattern="'*.txt'",
                 exclude="'bar.txt'")
    results = run(capsys, runner, tmp_path)
    results.out.assert_has_run(runner, '/foo.txt')
    results.out.assert_has_run(runner, '/child/foo.txt')
    assert results.total == 2, results.out.text


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_filter_include_filenames(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    write_doctest(tmp_path, 'foo.txt')
    write_doctest(tmp_path, 'bar.txt')
    write_doctest(tmp_path, 'baz', 'bar.txt')
    write_config(tmp_path, runner,
                 filenames="['bar.txt']")
    results = run(capsys, runner, tmp_path)
    results.out.assert_has_run(runner, '/bar.txt')
    results.out.assert_has_run(runner, '/baz/bar.txt')
    assert results.total == 2, results.out.text


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_filter_globs(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    write_doctest(tmp_path, 'middle', 'interesting', 'foo.txt')
    write_doctest(tmp_path, 'middle', 'boring', 'bad1.txt')
    write_doctest(tmp_path, 'middle', 'boring', 'bad2.txt')
    write_config(tmp_path, runner,
                 patterns="['middle/**/*.txt']",
                 excludes="['**/boring/*.txt']")
    results = run(capsys, runner, tmp_path)
    results.out.assert_has_run(runner, '/middle/interesting/foo.txt')
    assert results.total == 1, results.out.text


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_filter_multiple_patterns(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    write_doctest(tmp_path, 'test.rst')
    write_doctest(tmp_path, 'test.txt')
    write_config(tmp_path, runner,
                 patterns="['*.rst', '*.txt']")
    results = run(capsys, runner, tmp_path)
    assert results.total == 2, results.out.text


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_skips(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    root = clone_functional_sample('skips', tmp_path)
    write_config(root, runner,
                 parsers="[PythonCodeBlockParser(), SkipParser(), DocTestParser()]",
                 patterns="['*.rst']")
    results = run(capsys, runner, root)
    assert results.total == 10, results.out.text
    assert results.failures == 0, results.out.text
    assert results.errors == 0, results.out.text


def clone_and_run_modules_tests(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    clone_functional_sample('modules', tmp_path)
    write_config(tmp_path, runner,
                 path="'./modules'",
                 parsers="[PythonCodeBlockParser(), DocTestParser()]",
                 patterns="['*.py']")
    results = run(capsys, runner, tmp_path)
    return results


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_modules(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    sys.path.append(str(tmp_path / 'modules'))
    results = clone_and_run_modules_tests(tmp_path, capsys, runner)
    assert results.total == 5, results.out.text
    assert results.failures == 0, results.out.text
    assert results.errors == 0, results.out.text


def test_modules_not_importable_pytest(tmp_path: Path, capsys: CaptureFixture[str]):
    # NB: no append to sys.path
    results = clone_and_run_modules_tests(tmp_path, capsys, PYTEST)
    compare(results.total, expected=5, suffix=results.out.text)
    compare(results.errors, expected=0, suffix=results.out.text)
    compare(results.failures, expected=5, suffix=results.out.text)
    out = results.out
    out.then_find('a.py line=3 column=1')
    out.then_find(f"ImportError: 'a' not importable from {tmp_path/'modules'/'a.py'} as:")
    out.then_find("ModuleNotFoundError: No module named 'a'")
    out.then_find('a.py line=7 column=1')
    out.then_find("ModuleNotFoundError: No module named 'a'")
    out.then_find('b.py line=2 column=1')
    out.then_find(f"ImportError: 'b' not importable from {tmp_path/'modules'/'b.py'} as:")
    out.then_find('b.py line=7 column=1')
    out.then_find("ModuleNotFoundError: No module named 'b'")
    out.then_find('b.py line=11 column=1')
    out.then_find("ModuleNotFoundError: No module named 'b'")


def test_modules_not_importable_unittest(tmp_path: Path, capsys: CaptureFixture[str]):
    # NB: no append to sys.path
    results = clone_and_run_modules_tests(tmp_path, capsys, UNITTEST)
    assert results.total == 5, results.out.text
    assert results.failures == 0, results.out.text
    assert results.errors == 5, results.out.text
    a_py = tmp_path/'modules'/'a.py'
    b_py = tmp_path/'modules'/'b.py'
    out = results.out
    out.then_find(f'ERROR: {a_py},line:3,column:1')
    out.then_find(f"ImportError: 'a' not importable from {tmp_path/'modules'/'a.py'} as:")
    out.then_find("ModuleNotFoundError: No module named 'a'")
    out.then_find(f'ERROR: {b_py},line:2,column:1')
    out.then_find(f"ImportError: 'b' not importable from {tmp_path/'modules'/'b.py'} as:")
    out.then_find("ModuleNotFoundError: No module named 'b'")


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_package_and_docs(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    root = clone_functional_sample('package_and_docs', tmp_path)
    write_config(root, runner,
                 patterns="['**/*.py', '**/*.rst']")
    sys.path.append(str((root / 'src')))
    results = run(capsys, runner, root)
    assert results.total == 7, results.out.text
    assert results.failures == 1, results.out.text
    assert results.errors == 0, results.out.text
    # output from the one expected failure!
    results.out.then_find('Expected:')
    results.out.then_find('good')
    results.out.then_find('Got:')
    results.out.then_find('bad')


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_multiple_sybils_process_all(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    write_doctest(tmp_path, 'test.rst')
    write_doctest(tmp_path, 'test.txt')
    config_template = """
    from sybil.parsers.rest import DocTestParser
    from sybil import Sybil
    
    sybil1 = Sybil(parsers=[DocTestParser()], pattern='test.*', name='a')
    sybil2 = Sybil(parsers=[DocTestParser()], pattern='test.*', name='b')
    
    {assigned_name} = (sybil1 + sybil2).{integration}()
    """
    write_config(tmp_path, runner, template=config_template)
    results = run(capsys, runner, tmp_path)
    compare(results.total, expected=4, suffix=results.out.text)
    # make sure the name is used:
    results.out.assert_has_run(runner, 'test.rst', sybil='a')
    results.out.assert_has_run(runner, 'test.rst', sybil='b')
    results.out.assert_has_run(runner, 'test.txt', sybil='a')
    results.out.assert_has_run(runner, 'test.txt', sybil='b')


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_multiple_sybils_process_one_each(tmp_path: Path, capsys: CaptureFixture[str], runner: str):
    write_doctest(tmp_path, 'test.rst')
    write_doctest(tmp_path, 'test.txt')
    config_template = """
    from sybil.parsers.rest import DocTestParser
    from sybil import Sybil

    rst = Sybil(parsers=[DocTestParser()], pattern='*.rst')
    txt = Sybil(parsers=[DocTestParser()], pattern='*.txt')

    {assigned_name} = (rst + txt).{integration}()
    """
    write_config(tmp_path, runner, template=config_template)
    results = run(capsys, runner, tmp_path)
    compare(results.total, expected=2, suffix=results.out.text)


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_myst(capsys: CaptureFixture[str], runner: str):
    results = run(capsys, runner, functional_sample('myst'))
    out = results.out

    # Check all the tests are found:
    out.assert_has_run(runner, '/doctest.md', line=7)
    out.assert_has_run(runner, '/doctest.md', line=14)
    out.assert_has_run(runner, '/doctest.md', line=25)
    out.assert_has_run(runner, '/doctest.md', line=31)
    out.assert_has_run(runner, '/python.md', line=5)
    out.assert_has_run(runner, '/python.md', line=11)
    out.assert_has_run(runner, '/python.md', line=17)
    out.assert_has_run(runner, '/python.md', line=26)
    out.assert_has_run(runner, '/python.md', line=41)
    out.assert_has_run(runner, '/python.md', line=47)
    out.assert_has_run(runner, '/python.md', line=49)

    # unittest treats exceptions as errors rather than failures,
    # and they appear at the top of the output, hence the conditionals below.

    # Check counts:
    assert results.total == 4+7, results.out.text
    if runner == PYTEST:
        assert results.failures == 2 + 1, results.out.text
        assert results.errors == 0, results.out.text
    else:
        assert results.failures == 2 + 0, results.out.text
        assert results.errors == 1, results.out.text

    # Check error text:
    if runner == UNITTEST:
        out.then_find("Exception: boom!")
    out.then_find("doctest.md, line 25, column 1 did not evaluate as expected:")
    out.then_find("Expected:\n    3\nGot:\n    2\n")
    out.then_find("doctest.md, line 31, column 1 did not evaluate as expected:")
    out.then_find("Expected:\n    4\nGot:\n    2\n")
    if runner == PYTEST:
        out.then_find("Exception: boom!")


@pytest.mark.parametrize('runner', [PYTEST, UNITTEST])
def test_markdown(capsys: CaptureFixture[str], runner: str):
    results = run(capsys, runner, functional_sample('markdown'))
    out = results.out

    # Check all the tests are found:
    out.assert_has_run(runner, '/doctest.md', line=7)
    out.assert_has_run(runner, '/python.md', line=5)
    out.assert_has_run(runner, '/python.md', line=11)
    out.assert_has_run(runner, '/python.md', line=26)
    out.assert_has_run(runner, '/python.md', line=32)
    out.assert_has_run(runner, '/python.md', line=34)

    # unittest treats exceptions as errors rather than failures,
    # and they appear at the top of the output, hence the conditionals below.

    # Check counts:
    assert results.total == 1+5, results.out.text
    if runner == PYTEST:
        assert results.failures == 0 + 1, results.out.text
        assert results.errors == 0, results.out.text
    else:
        assert results.failures == 0 + 0, results.out.text
        assert results.errors == 1, results.out.text

    # Check error text:
    out.then_find("Exception: boom!")


def test_codeblock_with_protocol_then_doctest():
    sybil = Sybil([PythonCodeBlockParser(), DocTestParser()])
    check_path(sample_path('protocol-typing.rst'), sybil, expected=3)


def test_request_missing_fixtures(tmp_path: Path, capsys: CaptureFixture[str]):
    write_doctest(tmp_path, 'test.rst')
    config_template = """
    from sybil.parsers.rest import DocTestParser
    from sybil import Sybil

    sybil = Sybil(parsers=[DocTestParser()], pattern='*.rst', fixtures=['bad'])

    {assigned_name} = sybil.{integration}()
    """
    write_config(tmp_path, PYTEST, template=config_template)
    results = run(capsys, PYTEST, tmp_path)
    compare(results.total, expected=1, suffix=results.out.text)
    compare(results.failures, expected=1, suffix=results.out.text)
    compare(results.errors, expected=0, suffix=results.out.text)



from typing import Any, Callable, Dict, Optional
from unittest import TestCase as BaseTestCase, TestSuite
from unittest.loader import TestLoader

from sybil import Sybil
from sybil.example import Example


class TestCase(BaseTestCase):

    sybil: Sybil
    namespace: Dict[str, Any]

    def __init__(self, example: Example) -> None:
        BaseTestCase.__init__(self)
        self.example = example

    def runTest(self) -> None:
        self.example.evaluate()

    def id(self) -> str:
        return f'{self.example.path},{self.sybil.identify(self.example)}'

    __str__ = __repr__ = id

    @classmethod
    def setUpClass(cls) -> None:
        if cls.sybil.setup is not None:
            cls.sybil.setup(cls.namespace)

    @classmethod
    def tearDownClass(cls) -> None:
        if cls.sybil.teardown is not None:
            cls.sybil.teardown(cls.namespace)


def unittest_integration(
    *sybils: Sybil,
) -> Callable[[Optional[TestLoader], Optional[TestSuite], Optional[str]], TestSuite]:

    def load_tests(
        loader: Optional[TestLoader] = None,
        tests: Optional[TestSuite] = None,
        pattern: Optional[str] = None,
    ) -> TestSuite:
        suite = TestSuite()
        for sybil in sybils:
            for path in sorted(sybil.path.glob('**/*')):
                if path.is_file() and sybil.should_parse(path):
                    document = sybil.parse(path)

                    case = type(document.path, (TestCase, ), dict(
                        sybil=sybil, namespace=document.namespace,
                    ))

                    for example in document.examples():
                        suite.addTest(case(example))

        return suite

    return load_tests



import re
from typing import Optional, Dict, Iterable

from sybil import Document, Region
from sybil.parsers.abstract.lexers import BlockLexer
from sybil.parsers.markdown.lexers import RawFencedCodeBlockLexer
from sybil.parsers.rest.lexers import parse_options_and_source

INFO_PATTERN = (
    r'\{{(?P<directive>{directive})}} ?(?P<arguments>{arguments})$\n'
    r'(?P<options>(?:[ \t]*:[\w-]*:[^\n]*\n)+)?'
    r"([ \t]*---\n(?P<yaml_options>(?:.+\n)*)[ \t]*---\n)?"
)


def parse_yaml_options(lexed: Region) -> None:
    lexemes = lexed.lexemes
    yaml_options = lexemes.pop('yaml_options', None)
    if yaml_options is not None:
        # import here to avoid a dependency on PyYAML except where it's really needed:
        from yaml import safe_load
        options = safe_load(yaml_options)
        lexemes['options'].update(options)


class DirectiveLexer(RawFencedCodeBlockLexer):
    """
    A :class:`~sybil.typing.Lexer` for MyST directives such as:

    .. code-block:: markdown

        ```{directivename} arguments
        ---
        key1: val1
        key2: val2
        ---
        This is
        directive content
        ```

    The following lexemes are extracted:

    - ``directive`` as a  :class:`str`.
    - ``arguments`` as a :class:`str`.
    - ``source`` as a :class:`~sybil.Lexeme`.

    :param directive:
        a :class:`str` containing a regular expression pattern to match directive names.

    :param arguments:
        a :class:`str` containing a regular expression pattern to match directive arguments.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their
        values. Only mapped lexemes will be returned in any :class:`~sybil.Region` objects.

    """

    def __init__(
            self, directive: str, arguments: str = '.*', mapping: Optional[Dict[str, str]] = None
    ) -> None:
        super().__init__(
            info_pattern=re.compile(
                INFO_PATTERN.format(directive=directive, arguments=arguments),
                re.MULTILINE,
            ),
            mapping=mapping,
        )

    def __call__(self, document: Document) -> Iterable[Region]:
        for lexed in super().__call__(document):
            parse_options_and_source(lexed)
            parse_yaml_options(lexed)
            yield lexed


DIRECTIVE_IN_PERCENT_COMMENT_START = (
    r"^(?P<prefix>[ \t]*%)[ \t]*(?P<directive>{directive})(:[ \t]*(?P<arguments>{arguments}))?$\n"
)
DIRECTIVE_IN_PERCENT_COMMENT_END = '(?<=\n)(?!{prefix})'


class DirectiveInPercentCommentLexer(BlockLexer):
    """
    A :class:`~sybil.parsers.abstract.lexers.BlockLexer` for faux MyST directives in
    %-style Markdown comments such as:

    .. code-block:: markdown

        % not-really-a-directive: some-argument
        %
        %     Source here...

    It extracts the following lexemes:

    - ``directive`` as a  :class:`str`.
    - ``arguments`` as a :class:`str`.
    - ``source`` as a :class:`~sybil.Lexeme`.

    :param directive:
        a :class:`str` containing a regular expression pattern to match directive names.

    :param arguments:
        a :class:`str` containing a regular expression pattern to match directive arguments.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their values.
        Only mapped lexemes will be returned in any :class:`~sybil.Region` objects.
    """

    def __init__(
            self, directive: str, arguments: str = '.*', mapping: Optional[Dict[str, str]] = None
    ) -> None:
        super().__init__(
            start_pattern=re.compile(
                DIRECTIVE_IN_PERCENT_COMMENT_START.format(directive=directive, arguments=arguments),
                re.MULTILINE
            ),
            end_pattern_template=DIRECTIVE_IN_PERCENT_COMMENT_END,
            mapping=mapping,
        )





import re

from sybil.parsers.abstract import AbstractClearNamespaceParser
from .lexers import DirectiveInPercentCommentLexer
from ..markdown.lexers import DirectiveInHTMLCommentLexer


class ClearNamespaceParser(AbstractClearNamespaceParser):
    """
    A :any:`Parser` for :ref:`clear-namespace <myst-clear-namespace>` instructions.
    """

    def __init__(self) -> None:
        super().__init__([
            DirectiveInPercentCommentLexer('clear-namespace'),
            DirectiveInHTMLCommentLexer('clear-namespace'),
        ])



import ast
import re
from ast import AsyncFunctionDef, FunctionDef, ClassDef, Module, Expr, Constant, Str
from bisect import bisect
from io import open
from itertools import chain
from pathlib import Path
from typing import Any, Dict, cast
from typing import List, Iterator, Pattern, Tuple, Match

from .example import Example, SybilFailure, NotEvaluated
from .exceptions import LexingException
from .python import import_path
from .region import Region
from .text import LineNumberOffsets
from .typing import Parser, Evaluator


class Document:
    """
    This is Sybil's representation of a documentation source file.
    It will be instantiated by Sybil and provided to each parser in turn.

    Different types of document can be handled by subclassing to provide the
    required :any:`evaluation <push_evaluator>`. The required file extensions, such as ``'.py'``,
    can then be mapped to these subclasses using :class:`Sybil's <Sybil>`
    ``document_types`` parameter.
    """

    def __init__(self, text: str, path: str) -> None:
        #: This is the text of the documentation source file.
        self.text: str = text
        #: This is the absolute path of the documentation source file.
        self.path: str = path
        self.end: int = len(text)
        self.regions: List[Tuple[int, Region]] = []
        #: This dictionary is the namespace in which all examples parsed from
        #: this document will be evaluated.
        self.namespace: Dict[str, Any] = {}
        self.evaluators: list[Evaluator] = []

    @classmethod
    def parse(cls, path: str, *parsers: Parser, encoding: str = 'utf-8') -> 'Document':
        """
        Read the text from the supplied path and parse it into a document
        using the supplied parsers.
        """
        with open(path, encoding=encoding) as source:
            text = source.read()
        document = cls(text, path)
        for parser in parsers:
            for region in parser(document):
                document.add(region)
        return document

    def line_column(self, position: int) -> str:
        """
        Return a line and column location in this document based on a character
        position.
        """
        line = self.text.count('\n', 0, position)+1
        col = position - self.text.rfind('\n', 0, position)
        return 'line {}, column {}'.format(line, col)

    def region_details(self, region: Region) -> str:
        return '{!r} from {} to {}'.format(
            region,
            self.line_column(region.start),
            self.line_column(region.end)
        )

    def raise_overlap(self, *regions: Region) -> None:
        reprs = []
        for region in regions:
            reprs.append(self.region_details(region))
        raise ValueError('{} overlaps {}'.format(*reprs))

    def add(self, region: Region) -> None:
        if region.start < 0:
            raise ValueError('{} is before start of document'.format(
                self.region_details(region)
            ))
        if region.end > self.end:
            raise ValueError('{} goes beyond end of document'.format(
                self.region_details(region)
            ))
        entry = (region.start, region)
        index = bisect(self.regions, entry)
        if index > 0:
            previous = self.regions[index-1][1]
            if previous.end > region.start:
                self.raise_overlap(previous, region)
        if index < len(self.regions):
            next = self.regions[index][1]
            if next.start < region.end:
                self.raise_overlap(region, next)
        self.regions.insert(index, entry)

    def examples(self) -> Iterator[Example]:
        """
        Return the :term:`examples <example>` contained within this document.
        """
        line = 1
        place = 0
        for _, region in self.regions:
            line += self.text.count('\n', place, region.start)
            line_start = self.text.rfind('\n', place, region.start)
            place = region.start
            yield Example(self,
                          line, region.start-line_start,
                          region, self.namespace)

    def __iter__(self) -> Iterator[Example]:
        return self.examples()

    def find_region_sources(
        self, start_pattern: Pattern[str], end_pattern: Pattern[str]
    ) -> Iterator[Tuple[Match[str], Match[str], str]]:
        """
        This helper method can be used to extract source text
        for regions based on the two :ref:`regular expressions <re-objects>`
        provided.
        
        It will yield a tuple of ``(start_match, end_match, source)`` for each 
        occurrence of ``start_pattern`` in the document's 
        :attr:`~Document.text` that is followed by an
        occurrence of ``end_pattern``.
        The matches will be provided as :ref:`match objects <match-objects>`,
        while the source is provided as a string.
        """
        for start_match in re.finditer(start_pattern, self.text):
            source_start = start_match.end()
            end_match = end_pattern.search(self.text, source_start)
            if end_match is None:
                raise LexingException(
                    f'Could not match {end_pattern.pattern!r} in {self.path}:\n'
                    f'{self.text[source_start:]!r}'
                )
            source_end = end_match.start()
            source = self.text[source_start:source_end]
            yield start_match, end_match, source

    def push_evaluator(self, evaluator: Evaluator) -> None:
        """
        Push an :any:`Evaluator` onto this document's stack of evaluators
        if it is not already in that stack.

        When evaluating an :any:`Example`, any evaluators in the stack will be tried in order,
        starting with the most recently pushed. If an evaluator raises a :any:`NotEvaluated`
        exception, then the next evaluator in the stack will be attempted.

        If the stack is empty or all evaluators present raise :any:`NotEvaluated`, then the
        example's evaluator will be used. This is the most common case!
        """
        if evaluator not in self.evaluators:
            self.evaluators.append(evaluator)

    def pop_evaluator(self, evaluator: Evaluator) -> None:
        """
        Pop an :any:`Evaluator` off this document's stack of evaluators.
        If it is not present in that stack, the method does nothing.
        """
        if evaluator in self.evaluators:
            self.evaluators.remove(evaluator)

    def evaluate(self, example: Example, evaluator: Evaluator) -> None:

        __tracebackhide__ = True

        for current_evaluator in chain(reversed(self.evaluators), (evaluator,)):
            try:
                result = current_evaluator(example)
            except NotEvaluated:
                continue
            else:
                if result:
                    raise SybilFailure(example, result if isinstance(result, str) else repr(result))
                else:
                    break
        else:
            raise SybilFailure(example, f'{evaluator!r} should not raise NotEvaluated()')


DOCSTRING_PUNCTUATION = re.compile('[rf]?(["\']{3}|["\'])')


class PythonDocument(Document):
    """
    A :class:`~sybil.Document` type that imports the document's source
    file as a Python module, making names within it available in the document's
    :attr:`~sybil.Document.namespace`.
    """

    def __init__(self, text: str, path: str) -> None:
        super().__init__(text, path)
        self.push_evaluator(self.import_document)

    def import_document(self, example: Example) -> None:
        """
        Imports the document's source file as a Python module when the first
        :class:`~sybil.example.Example` from it is evaluated.
        """
        module = import_path(Path(self.path))
        self.namespace.update(module.__dict__)
        self.pop_evaluator(self.import_document)
        raise NotEvaluated()


class PythonDocStringDocument(PythonDocument):
    """
    A :class:`~sybil.document.PythonDocument` subclass that only considers the text of
    docstrings in the document's source.
    """

    @staticmethod
    def extract_docstrings(python_source_code: str) -> Iterator[Tuple[int, int, str]]:
        line_offsets = LineNumberOffsets(python_source_code)
        for node in ast.walk(ast.parse(python_source_code)):
            if not isinstance(node, (AsyncFunctionDef, FunctionDef, ClassDef, Module)):
                continue
            if not (node.body and isinstance(node.body[0], Expr)):
                continue
            docstring = node.body[0].value
            if isinstance(docstring, Str):
                text = docstring.s
            else:
                continue
            node_start = line_offsets.get(docstring.lineno-1, docstring.col_offset)
            end_lineno = docstring.end_lineno or 1
            end_col_offset = docstring.end_col_offset or 0
            node_end = line_offsets.get(end_lineno-1, end_col_offset)
            punc = DOCSTRING_PUNCTUATION.match(python_source_code, node_start, node_end)
            punc_size = len(punc.group(1))
            start = punc.end()
            end = node_end - punc_size
            yield start, end, text

    @classmethod
    def parse(cls, path: str, *parsers: Parser, encoding: str = 'utf-8') -> 'Document':
        """
        Read the text from the supplied path to a Python source file and parse any docstrings
        it contains into a document using the supplied parsers.
        """
        with open(path, encoding=encoding) as source:
            document = cls(source.read(), path)
            for start, end, text in cls.extract_docstrings(document.text):
                docstring_document = cls(text, path)
                for parser in parsers:
                    for region in parser(docstring_document):
                        region.start += start
                        region.end += start
                        document.add(region)
        return document



from typing import Callable, TYPE_CHECKING, Iterable, Optional, Any, Dict

if TYPE_CHECKING:
    import sybil


#: The signature for an :term:`evaluator`.
Evaluator = Callable[['sybil.Example'], Optional[str]]

#: The signature for a :term:`lexer`.
#: Lexers must not set :attr:`~sybil.Region.parsed` or :attr:`~sybil.Region.evaluator`
#: on the :class:`~sybil.Region` instances they return.
Lexer = Callable[['sybil.Document'], Iterable['sybil.Region']]

#: The signature for a :term:`parser`.
Parser = Callable[['sybil.Document'], Iterable['sybil.Region']]

# In the future, this could likely be a TypedDict when Python 3.8 is the minimum supported version
#: Mappings used to store lexemes for a :class:`~sybil.Region`.
LexemeMapping = Dict[str, Any]




from typing import Iterable, Sequence

from sybil import Document, Region, Example
from sybil.parsers.abstract.lexers import LexerCollection
from sybil.typing import Lexer


class AbstractClearNamespaceParser:
    """
    An abstract parser for clearing the :class:`~sybil.Document.namespace`.
    """

    def __init__(self, lexers: Sequence[Lexer]) -> None:
        self.lexers = LexerCollection(lexers)

    @staticmethod
    def evaluate(example: Example) -> None:
        example.document.namespace.clear()

    def __call__(self, document: Document) -> Iterable[Region]:
        for lexed in self.lexers(document):
            yield Region(lexed.start, lexed.end, lexed.lexemes['source'], self.evaluate)



from pathlib import Path

from testfixtures import compare

from sybil.parsers.markdown.lexers import FencedCodeBlockLexer, DirectiveInHTMLCommentLexer
from sybil.parsers.myst.lexers import (
    DirectiveLexer,
    DirectiveInPercentCommentLexer
)
from sybil.region import Region
from .helpers import lex, sample_path, lex_text, check_lexed_regions, check_lexed_text_regions


def test_fenced_code_block():
    lexer = FencedCodeBlockLexer('py?thon')
    check_lexed_regions('myst-lexers.md', lexer, expected=[
        Region(36, 59, lexemes={'language': 'python', 'source': '>>> 1+1\n2\n'}),
        Region(1137, 1173, lexemes={'language': 'pthon', 'source': 'assert 1 + 1 == 2\n'}),
    ])


def test_fenced_code_block_with_mapping():
    lexer = FencedCodeBlockLexer('python', mapping={'source': 'body'})
    check_lexed_regions('myst-lexers.md', lexer, expected=[
        Region(36, 59, lexemes={'body': '>>> 1+1\n2\n'})
    ])


def test_myst_directives():
    lexer = DirectiveLexer(directive='[^}]+')
    check_lexed_regions('myst-lexers.md', lexer, expected=[
        Region(110, 148, lexemes={
            'directive': 'code-block',
            'arguments': 'python',
            'options': {},
            'source': '>>> 1 + 1\n3\n',
        }),
        Region(188, 276, lexemes={
            'directive': 'directivename',
            'arguments': 'arguments',
            'options': {'key1': 'val1', 'key2': 'val2'},
            'source': 'This is\ndirective content\n',
        }),
        Region(330, 381, lexemes={
            'directive': 'eval-rst',
            'arguments': '',
            'options': {},
            'source': '.. doctest::\n\n    >>> 1 + 1\n    4\n',
        }),
        Region(1398, 1479, lexemes={
            'directive': 'foo',
            'arguments': 'bar',
            'options': {'key1': 'val1'},
            'source': 'This, too, is a directive content\n',
        }),
    ])


def test_examples_from_parsing_tests():
    lexer = DirectiveLexer(directive='code-block', arguments='python')
    compare(lex('myst-codeblock.md', lexer), expected=[
        Region(99, 154, lexemes={
            'directive': 'code-block',
            'arguments': 'python',
            'options': {},
            'source': "raise Exception('boom!')\n",
        }),
        Region(701, 753, lexemes={
            'directive': 'code-block',
            'arguments': 'python',
            'options': {},
            'source': 'define_this = 1\n\n',
        }),
    ])


def test_myst_directives_with_mapping():
    lexer = DirectiveLexer(directive='directivename', arguments='.*', mapping={'arguments': 'foo'})
    compare(lex('myst-lexers.md', lexer), expected=[
        Region(188, 276, lexemes={'foo': 'arguments', 'options': {}}),
    ])


def test_myst_percent_comment_invisible_directive():
    lexer = DirectiveInPercentCommentLexer(
        directive='(invisible-)?code(-block)?'
    )
    compare(lex('myst-lexers.md', lexer), expected=[
        Region(449, 504, lexemes={
            'directive': 'invisible-code-block', 'arguments': 'python',
            'source': '\nb = 5\n\n...etc...\n',
        }),
        Region(584, 652, lexemes={
            'directive': 'code-block', 'arguments': 'py',
            'source': '\nb = 6\n...etc...\n\n',
        }),
    ])


def test_myst_percent_comment_invisible_directive_mapping():
    lexer = DirectiveInPercentCommentLexer(
        directive='inv[^:]+', arguments='python', mapping={'arguments': 'language'}
    )
    compare(lex('myst-lexers.md', lexer), expected=[
        Region(449, 504, lexemes={'language': 'python'}),
    ])


def test_myst_html_comment_invisible_directive():
    lexer = DirectiveInHTMLCommentLexer(
        directive='(invisible-)?code(-block)?'
    )
    compare(lex('myst-lexers.md', lexer), show_whitespace=True, expected=[
        Region(702, 827, lexemes={
            'directive': 'invisible-code-block', 'arguments': 'python',
            'source': (
                'def foo():\n'
                '   return 42\n\n'
                'meaning_of_life = 42\n\n'
                'assert foo() == meaning_of_life()\n'
            ),
        }),
        Region(912, 1015, lexemes={
            'directive': 'code-block', 'arguments': 'python',
            'source': (
                '\n'
                'blank line above ^^\n'
                '\n'
                'blank line below:\n'
                '\n'
            ),
        }),
        Region(1229, 1332, lexemes={
            'directive': 'invisible-code', 'arguments': 'py',
            'source': (
                '\n'
                'blank line above ^^\n'
                '\n'
                'blank line below:\n'
                '\n'
            ),
        }),
    ])


def test_myst_html_comment_invisible_skip_directive():
    lexer = DirectiveInHTMLCommentLexer(directive='skip')
    compare(lex('myst-lexers.md', lexer), show_whitespace=True, expected=[
        Region(1482, 1498, lexemes={
            'directive': 'skip',
            'arguments': 'next',
            'source': '',
        }),
        Region(1503, 1562, lexemes={
            'directive': 'skip',
            'arguments': 'start if("some stuff here", reason=\'Something\')' ,
            'source': '',
        }),
        Region(1567, 1584, lexemes={
            'directive': 'skip',
            'arguments': 'and',
            'source': '',
        }),
        Region(1589, 1647, lexemes={
            'directive': 'skip',
            'arguments': 'end',
            'source': '\n\n\nOther stuff here just gets ignored\n\n',
        }),
        Region(1652, 1672, lexemes={
            'directive': 'skip',
            'arguments': 'also',
            'source': '',
        }),
    ])


def test_myst_html_comment_invisible_clear_directive():
    lexer = DirectiveInHTMLCommentLexer('clear-namespace')
    compare(lex('myst-lexers.md', lexer), show_whitespace=True, expected=[
        Region(1678, 1699, lexemes={
            'directive': 'clear-namespace',
            'arguments': '',
            'source': '',
        }),
    ])


def test_lexing_directives():
    lexer = DirectiveLexer('[^}]+')
    compare(lex('myst-lexing-directives.md', lexer), expected=[
        Region(55, 236, lexemes={
            'directive': 'note',
            'arguments': 'This is a note admonition.',
            'options': {},
            'source': ('This is the second line of the first paragraph.\n'
                       '\n'
                       '- The note contains all indented body elements\n'
                       '  following.\n'
                       '- It includes this bullet list.\n'),
        }),
        Region(238, 320, lexemes={
            'directive': 'admonition',
            'arguments': 'And, by the way...',
            'options': {},
            'source': 'You can make up your own admonition too.\n',
        }),
        Region(322, 386, lexemes={
            'directive': 'sample',
            'arguments': '',
            'options': {},
            'source': 'This directive has no arguments, just a body.\n',
        }),
        Region(455, 481, lexemes={
            'directive': 'image',
            'arguments': 'picture.png',
            'options': {},
            'source': '',
        }),
        Region(483, 595, lexemes={
            'directive': 'image',
            'arguments': 'picture.jpeg',
            'options': {
                'height': '100px',
                'width': '200 px',
                'scale': '50 %',
                'alt': 'alternate text',
                'align': 'right',
            },
            'source': '',
        }),
        Region(597, 1314, lexemes={
            'directive': 'figure',
            'arguments': 'picture.png',
            'options': {
                'alt': 'map to buried treasure',
                'scale': '50 %',
            },
            'source': ('This is the caption of the figure (a simple paragraph).\n'
                       '\n'
                       'The legend consists of all elements after the caption.  In this\n'
                       'case, the legend consists of this paragraph and the following\n'
                       'table:\n'
                       '\n'
                       '+-----------------------+-----------------------+\n'
                       '| Symbol                | Meaning               |\n'
                       '+=======================+=======================+\n'
                       '| .. image:: tent.png   | Campground            |\n'
                       '+-----------------------+-----------------------+\n'
                       '| .. image:: waves.png  | Lake                  |\n'
                       '+-----------------------+-----------------------+\n'
                       '| .. image:: peak.png   | Mountain              |\n'
                       '+-----------------------+-----------------------+\n'
                       '\n')
        }),
        Region(1317, 1452, lexemes={
            'directive': 'topic',
            'arguments': 'Topic Title',
            'options': {},
            'source': ('Subsequent indented lines comprise\n'
                       'the body of the topic, and are\n'
                       'interpreted as body elements.\n')
        }),
        Region(1506, 1592, lexemes={
            'directive': 'topic',
            'arguments': 'example.cfg',
            'options': {'class': 'read-file'},
            'source': '::\n\n  [A Section]\n  dir = frob\n'
        }),
        Region(1612, 1804, lexemes={
            'directive': 'sidebar',
            'arguments': 'Optional Sidebar Title',
            'options': {'subtitle': 'Optional Sidebar Subtitle'},
            'source': ('Subsequent indented lines comprise\n'
                       'the body of the sidebar, and are\n'
                       'interpreted as body elements.\n')
        }),
        Region(1807, 2006, lexemes={
            'directive': 'code-block',
            'arguments': 'python',
            'options': {'lineno-start': 10, 'emphasize-lines': '1, 3',
                        'caption': 'This is my\nmulti-line caption. It is *pretty nifty* ;-)\n'},
            'source': "a = 2\nprint('my 1st line')\nprint(f'my {a}nd line')\n",
        }),
        Region(2008, 2213, lexemes={
            'directive': 'eval-rst',
            'arguments': '',
            'options': {},
            'source': (
                '.. figure:: img/fun-fish.png\n'
                '  :width: 100px\n'
                '  :name: rst-fun-fish\n'
                '\n'
                '  Party time!\n'
                '\n'
                'A reference from inside: :ref:`rst-fun-fish`\n'
                '\n'
                'A reference from outside: :ref:`syntax/directives/parsing`\n'
            )
        }),
    ])


def test_directive_no_trailing_newline():
    lexer = DirectiveLexer(directive='toctree')
    text = Path(sample_path('myst-directive-no-trailing-newline.md')).read_text().rstrip('\n')
    compare(lex_text(text, lexer), expected=[
        Region(16, 67, lexemes={
            'directive': 'toctree',
            'arguments': '',
            'options': {'maxdepth': '1'},
            'source': 'flask\npyramid\ncustom\n',
        }),
    ])


def test_directive_nested():
    lexer = DirectiveLexer(directive='.+')
    text = Path(sample_path('myst-directive-nested.md')).read_text().rstrip('\n')
    check_lexed_text_regions(text, lexer, expected=[
        Region(54, 97, lexemes={
            'directive': 'warning',
            'arguments': '',
            'options': {},
            'source': "Here's my warning\n",
        }),
        Region(146, 222, lexemes={
            'directive': 'warning',
            'arguments': '',
            'options': {},
            'source': "Here's my raw text warning that isn't parsed...\n",
        }),
        Region(0, 227, lexemes={
            'directive': 'note',
            'arguments': '',
            'options': {},
            'source': ('The warning block will be properly-parsed\n'
                       '\n'
                       '   ```{warning}\n'
                       "   Here's my warning\n"
                       '   ```\n'
                       '\n'
                       'But the next block will be parsed as raw text\n'
                       '\n'
                       '    ```{warning}\n'
                       "    Here's my raw text warning that isn't parsed...\n"
                       '    ```\n'),
        }),
    ])


def test_complicated_nesting():
    lexer = DirectiveLexer(directive='.+')
    check_lexed_regions('myst-complicated-nesting.md', lexer, expected=[
        Region(37, 79, lexemes={
            'directive': 'py:module',
            'arguments': 'bytewax.connectors.demo',
            'options': {},
            'source': "",
        }),
        Region(81, 160, lexemes={
            'directive': 'autodoc2-docstring',
            'arguments': 'bytewax.connectors.demo',
            'options': {'parser': 'myst', 'allowtitles': ''},
            'source': "",
        }),
        Region(248, 315, lexemes={
            'directive': 'autodoc2-docstring',
            'arguments': 'bytewax.connectors.demo.X',
            'options': {'parser': 'myst'},
            'source': "",
        }),
        Region(171, 321, lexemes={
            'directive': 'py:data',
            'arguments': 'X',
            'options': {'canonical': 'bytewax.connectors.demo.X', 'type': 'typing.TypeVar'},
            'source': "```{autodoc2-docstring} bytewax.connectors.demo.X\n"
                      ":parser: myst\n"
                      "```\n\n",
        }),
        Region(789, 873, lexemes={
            'directive': 'autodoc2-docstring',
            'arguments': 'bytewax.connectors.demo.RandomMetricSource',
            'options': {'parser': 'myst'},
            'source': "",
        }),
        Region(875, 905, lexemes={
            'directive': 'rubric',
            'arguments': 'Initialization',
            'options': {},
            'source': "",
        }),
        Region(907, 1000, lexemes={
            'directive': 'autodoc2-docstring',
            'arguments': 'bytewax.connectors.demo.RandomMetricSource.__init__',
            'options': {'parser': 'myst'},
            'source': "",
        }),
        Region(1002, 1122, lexemes={
            'directive': 'py:method',
            'arguments': 'list_parts() -> typing.List[str]',
            'options': {'canonical': 'bytewax.connectors.demo.RandomMetricSource.list_parts'},
            'source': "",
        }),
        Region(1124, 1336, lexemes={
            'directive': 'py:method',
            'arguments': 'build_part(now: datetime.datetime, for_part: str, resume_state: '
                         'typing.Optional[bytewax.connectors.demo._RandomMetricState])',
            'options': {'canonical': 'bytewax.connectors.demo.RandomMetricSource.build_part'},
            'source': "",
        }),
        Region(336, 1343, lexemes={
            'directive': 'py:class',
            'arguments': 'RandomMetricSource(metric_name: str, interval: datetime.timedelta = '
                         'timedelta(seconds=0.7), count: int = sys.maxsize, next_random: '
                         'typing.Callable[[], float] = lambda: random.randrange(0, 10))',
            'options': {'canonical': 'bytewax.connectors.demo.RandomMetricSource'},
            'source': ":Bases:\n"
                      "    - {py:obj}`~bytewax.inputs.FixedPartitionedSource``[`{py:obj}`"
                      "~typing.Tuple``[`{py:obj}`~str``, `{py:obj}`~float``],"
                      " `{py:obj}`~bytewax.connectors.demo._RandomMetricState``]`\n\n"
                      "```{autodoc2-docstring} bytewax.connectors.demo.RandomMetricSource\n"
                      ":parser: myst\n"
                      "```\n\n"
                      "```{rubric} Initialization\n"
                      "```\n\n"
                      "```{autodoc2-docstring} "
                      "bytewax.connectors.demo.RandomMetricSource.__init__\n"
                      ":parser: myst\n"
                      "```\n\n"
                      "````{py:method} list_parts() -> typing.List[str]\n"
                      ":canonical: bytewax.connectors.demo.RandomMetricSource.list_parts\n\n"
                      "````\n\n"
                      "````{py:method} build_part(now: datetime.datetime, for_part: str, "
                      "resume_state: typing.Optional[bytewax.connectors.demo._RandomMetricState])\n"
                      ":canonical: bytewax.connectors.demo.RandomMetricSource.build_part\n\n"
                      "````\n\n",
        }),
    ])



import re

import pytest
from testfixtures import ShouldRaise, compare
from testfixtures.comparison import compare_text, compare_dict

from sybil import Lexeme
from sybil.exceptions import LexingException
from sybil.parsers.abstract.lexers import BlockLexer
from .helpers import lex, sample_path


def test_examples_from_parsing_tests():
    lexer = BlockLexer(start_pattern=re.compile('START'), end_pattern_template='END')
    path = sample_path('lexing-fail.txt')
    with ShouldRaise(LexingException(f"Could not match 'END' in {path}:\n'\\nEDN\\n'")):
        lex('lexing-fail.txt', lexer)


def test_indented_block():
    lexer = BlockLexer(
        start_pattern=re.compile('(?P<prefix> *)START\n'),
        end_pattern_template=' {{{len_prefix}}}END'
    )
    region, = lex('lexing-indented-block.txt', lexer)
    compare(
        region.lexemes,
        expected={
            'source': Lexeme(
                "".join(
                    (
                        "\n",
                        "line 1\n",
                        "\n",
                        "  line 2\n",
                        "\n",
                    )
                ),
                offset=0,
                line_offset=0,
            ),
        }
    )


class TestLexemeStripping:

    @staticmethod
    def compare_lexeme(expected: Lexeme, actual: Lexeme):

        def _compare(x, y, context):
            if str(x) != str(y):
                return compare_text(x, y, context)
            return compare_dict(x.__dict__, y.__dict__, context)

        compare(expected=expected, actual=actual, comparers={Lexeme: _compare}, strict=True)

    @pytest.mark.parametrize("actual,message", [
        (Lexeme('bar', 10, 1), "'foo' (expected) != 'bar' (actual)"),
        (Lexeme('foo', 11, 1), "'offset': 10 (expected) != 11"),
        (Lexeme('foo', 10, 2), "'line_offset': 1 (expected) != 2 (actual)"),
        (Lexeme('foo', 12, 3), "line_offset': 1 (expected) != 3 (actual)\n"
                               "'offset': 10 (expected) != 12 (actual)"),
    ])
    def test_not_equal(self, actual: Lexeme, message: str):
        with ShouldRaise(AssertionError) as s:
            self.compare_lexeme(Lexeme('foo', 10, 1), actual)
        assert message in str(s.raised)

    def test_strip_no_newlines_present(self):
        self.compare_lexeme(
            actual=Lexeme('  foo  \n', 10, 1).strip_leading_newlines(),
            expected=Lexeme('  foo  \n', 10, 1)
        )

    def test_strip_one_newline_present(self):
        self.compare_lexeme(
            actual=Lexeme('\n  foo  \n', 10, 1).strip_leading_newlines(),
            expected=Lexeme('  foo  \n', 11, 2)
        )

    def test_strip_multiple_newlines_present(self):
        self.compare_lexeme(
            actual=Lexeme('\n\n\n  foo  \n', 13, 3).strip_leading_newlines(),
            expected=Lexeme('  foo  \n', 16, 6)
        )

    def test_strip_newlines_and_spaces_present(self):
        self.compare_lexeme(
            actual=Lexeme(' \n \n  foo  \n', 10, 1).strip_leading_newlines(),
            expected=Lexeme(' \n \n  foo  \n', 10, 1)
        )



import re
import textwrap
from itertools import chain
from typing import Optional, Dict, Iterable, Pattern, List

from sybil import Document
from sybil.exceptions import LexingException
from sybil.region import Lexeme, Region
from sybil.typing import Lexer


class LexerCollection(List[Lexer]):

    def __call__(self, document: Document) -> Iterable[Region]:
        return chain(*(lexer(document) for lexer in self))


class BlockLexer:
    """
    This is a base class useful for any :any:`Lexer` that must handle block-style languages
    such as ReStructured Text or MarkDown.

    It yields a sequence of :class:`~sybil.Region` objects for each case where the
    ``start_pattern`` matches. A ``source`` :class:`~sybil.Lexeme` is created from the text between
    the end of the start pattern and the start of the end pattern.

    :param start_pattern:
        This is used to match the start of the block. Any named groups will be returned
        in the :attr:`~sybil.Region.lexemes` :class:`dict` of resulting
        :class:`~sybil.Region` objects. If a ``prefix`` named group forms
        part of the match, this will be template substituted into the ``end_pattern_template``
        before it is compiled.

    :param end_pattern_template:
        This is used to match the end of any block found by the ``start_pattern``.
        It is templated with any ``prefix`` group from the ``start_pattern`` :class:`~typing.Match`
        and ``len_prefix``, the length of that prefix, before being compiled into a
        :class:`~typing.Pattern`.

    :param mapping:
        If provided, this is used to rename lexemes from the keys in the mapping to their values.
        Only mapped lexemes will be returned in any :class:`~sybil.Region` objects.
    """

    def __init__(
            self,
            start_pattern: Pattern[str],
            end_pattern_template: str,
            mapping: Optional[Dict[str, str]] = None,
    ) -> None:
        self.start_pattern = start_pattern
        self.end_pattern_template = end_pattern_template
        self.mapping = mapping

    def __call__(self, document: Document) -> Iterable[Region]:
        for start_match in re.finditer(self.start_pattern, document.text):
            source_start = start_match.end()
            lexemes = start_match.groupdict()
            prefix = lexemes.pop('prefix', '')
            end_pattern = re.compile(self.end_pattern_template.format(
                prefix=prefix, len_prefix=len(prefix)
            ))
            end_match = end_pattern.search(document.text, source_start)
            if end_match is None:
                raise LexingException(
                    f'Could not match {end_pattern.pattern!r} in {document.path}:\n'
                    f'{document.text[source_start:]!r}'
                )
            source_end = end_match.start()
            source = document.text[source_start:source_end]
            lexemes['source'] = Lexeme(
                strip_prefix(source, prefix),
                offset=source_start-start_match.start(),
                line_offset=start_match.group(0).count('\n')-1
            )
            if self.mapping:
                lexemes = {dest: lexemes[source] for source, dest in self.mapping.items()}
            yield Region(start_match.start(), source_end, lexemes=lexemes)


def strip_prefix(text: str, prefix: str) -> str:
    lines = text.splitlines(keepends=True)
    prefix_length = len(prefix)
    return textwrap.dedent(''.join(line[prefix_length:] or line[-1] for line in lines))



from __future__ import absolute_import

import os
from inspect import getsourcefile
from os.path import abspath
from pathlib import Path
from typing import Callable, Union, Tuple, Optional, Sequence, List

import pytest
from _pytest import fixtures
from _pytest._code.code import TerminalRepr, Traceback, ExceptionInfo
from _pytest._io import TerminalWriter
from _pytest.fixtures import FuncFixtureInfo
from _pytest.main import Session
from _pytest.nodes import Collector
from _pytest.python import Module

from sybil import example as example_module, Sybil, Document
from sybil.example import Example
from sybil.example import SybilFailure

PYTEST_VERSION = tuple(int(i) for i in pytest.__version__.split('.'))

example_module_path = abspath(getsourcefile(example_module))


class SybilFailureRepr(TerminalRepr):

    def __init__(self, item: 'SybilItem', message: str) -> None:
        self.item = item
        self.message = message

    def toterminal(self, tw: TerminalWriter) -> None:
        tw.line()
        for line in self.message.splitlines():
            tw.line(line)
        tw.line()
        tw.write(self.item.parent.name, bold=True, red=True)
        tw.line(":%s: SybilFailure" % self.item.example.line)


class SybilItem(pytest.Item):

    obj = None

    def __init__(self, parent, sybil, example: Example) -> None:
        super(SybilItem, self).__init__(sybil.identify(example), parent)
        self.example = example
        self.request_fixtures(sybil.fixtures)

    def request_fixtures(self, names):
        # pytest fixtures dance:
        fm = self.session._fixturemanager
        closure = fm.getfixtureclosure(initialnames=names, parentnode=self, ignore_args=set())
        names_closure, arg2fixturedefs = closure
        fixtureinfo = FuncFixtureInfo(argnames=names, initialnames=names, names_closure=names_closure, name2fixturedefs=arg2fixturedefs)
        self._fixtureinfo = fixtureinfo
        self.funcargs = {}
        self._request = fixtures.TopRequest(pyfuncitem=self, _ispytest=True)
        self.fixturenames = names_closure

    def reportinfo(self) -> Tuple[Union["os.PathLike[str]", str], Optional[int], str]:
        info = '%s line=%i column=%i' % (
            self.path.name, self.example.line, self.example.column
        )
        return self.example.path, self.example.line, info

    def getparent(self, cls):
        if cls is Module:
            return self.parent
        if cls is Session:
            return self.session

    def setup(self) -> None:
        self._request._fillfixtures()
        for name, fixture in self.funcargs.items():
            self.example.namespace[name] = fixture

    def runtest(self) -> None:
        self.example.evaluate()

    def _traceback_filter(self, excinfo: ExceptionInfo[BaseException]) -> Traceback:
        traceback = excinfo.traceback
        tb = traceback.cut(path=example_module_path)
        tb_entry = tb[1]
        if getattr(tb_entry, '_rawentry', None) is not None:
            traceback = Traceback(tb_entry._rawentry)
        return traceback

    def repr_failure(
        self,
        excinfo: ExceptionInfo[BaseException],
        style = None,
    ) -> Union[str, TerminalRepr]:
        if isinstance(excinfo.value, SybilFailure):
            return SybilFailureRepr(self, str(excinfo.value))
        return super().repr_failure(excinfo, style)


class SybilFile(pytest.File):

    def __init__(self, *, sybils: Sequence[Sybil], **kwargs) -> None:
        super(SybilFile, self).__init__(**kwargs)
        self.sybils: Sequence[Sybil] = sybils
        self.documents: List[Document] = []

    def collect(self):
        for sybil in self.sybils:
            document = sybil.parse(self.path)
            self.documents.append(document)
            for example in document.examples():
                yield SybilItem.from_parent(
                    self,
                    sybil=sybil,
                    example=example,
                )

    def setup(self) -> None:
        for sybil, document in zip(self.sybils, self.documents):
            if sybil.setup:
                sybil.setup(document.namespace)

    def teardown(self) -> None:
        for sybil, document in zip(self.sybils, self.documents):
            if sybil.teardown:
                sybil.teardown(document.namespace)


def pytest_integration(*sybils: Sybil) -> Callable[[Path, Collector], Optional[SybilFile]]:

    def pytest_collect_file(file_path: Path, parent: Collector) -> Optional[SybilFile]:
        active_sybils = [sybil for sybil in sybils if sybil.should_parse(file_path)]
        if active_sybils:
            return SybilFile.from_parent(parent, path=file_path, sybils=active_sybils)
        return None

    return pytest_collect_file



import __future__

import pytest
from testfixtures import compare

from sybil import Example, Sybil
from sybil.parsers.myst import PythonCodeBlockParser, CodeBlockParser
from .helpers import check_excinfo, parse, check_path, sample_path


def test_basic():
    examples, namespace = parse('myst-codeblock.md', PythonCodeBlockParser(), expected=7)
    namespace['y'] = namespace['z'] = 0
    assert examples[0].evaluate() is None
    assert namespace['y'] == 1
    assert namespace['z'] == 0
    with pytest.raises(Exception) as excinfo:
        examples[1].evaluate()
    compare(examples[1].parsed, expected="raise Exception('boom!')\n", show_whitespace=True)
    assert examples[1].line == 10
    check_excinfo(examples[1], excinfo, 'boom!', lineno=11)
    assert examples[2].evaluate() is None
    assert namespace['y'] == 1
    assert namespace['z'] == 1
    assert examples[3].evaluate() is None
    assert namespace['bin'] == b'x'
    assert namespace['uni'] == u'x'
    assert examples[4].evaluate() is None
    assert 'NoVars' in namespace
    assert examples[5].evaluate() is None
    assert namespace['define_this'] == 1
    assert examples[6].evaluate() is None
    assert 'YesVars' in namespace
    assert '__builtins__' not in namespace


def test_complicated_nesting():
    # This has no code blocks, but should still parse fine:
    parse('myst-complicated-nesting.md', PythonCodeBlockParser(), expected=0)


def test_doctest_at_end_of_fenced_codeblock():
    examples, namespace = parse('myst-codeblock-doctests-end-of-fenced-codeblocks.md',
                                PythonCodeBlockParser(), expected=2)
    assert examples[0].evaluate() is None
    assert examples[1].evaluate() is None
    assert namespace['b'] == 2


def test_other_language_composition_pass():

    def oh_hai(example):
        assert isinstance(example, Example)
        assert 'HAI' in example.parsed

    parser = CodeBlockParser(language='lolcode', evaluator=oh_hai)
    examples, namespace = parse('myst-codeblock.md', parser, expected=1)
    assert examples[0].evaluate() is None



def test_other_language_composition_fail():
    def oh_noez(example):
        if 'KTHXBYE' in example.parsed:
            raise ValueError('oh noez')

    parser = CodeBlockParser(language='lolcode', evaluator=oh_noez)
    examples, namespace = parse('myst-codeblock.md', parser, expected=1)
    with pytest.raises(ValueError):
        examples[0].evaluate()


def test_other_language_no_evaluator():
    parser = CodeBlockParser('foo')
    with pytest.raises(NotImplementedError):
        parser.evaluate(...)


class LolCodeCodeBlockParser(CodeBlockParser):

    language = 'lolcode'

    def evaluate(self, example: Example):
        if example.parsed != 'HAI\n':
            raise ValueError(repr(example.parsed))


def test_other_language_inheritance():
    examples, namespace = parse('myst-codeblock-lolcode.md', LolCodeCodeBlockParser(), expected=2)
    examples[0].evaluate()
    with pytest.raises(ValueError) as excinfo:
        examples[1].evaluate()
    assert str(excinfo.value) == "'KTHXBYE\\n'"


class IgnoringPythonCodeBlockParser(CodeBlockParser):

    def __call__(self, document):
        for region in super().__call__(document):
            options = region.lexemes.get('options')
            if options and 'ignore' in options:
                continue
            yield region


class IgnoringCodeBlockParser(PythonCodeBlockParser):

    codeblock_parser_class = IgnoringPythonCodeBlockParser


def test_other_functionality_inheritance():
    examples, namespace = parse(
        'myst-codeblock-subclassing.md', IgnoringCodeBlockParser(), expected=2
    )
    examples[0].evaluate()
    examples[1].evaluate()


def future_import_checks(*future_imports):
    parser = PythonCodeBlockParser(future_imports)
    examples, namespace = parse('myst-codeblock-future-imports.md', parser, expected=3)
    with pytest.raises(Exception) as excinfo:
        examples[0].evaluate()
    # check the line number of the first block, which is the hardest case:
    check_excinfo(examples[0], excinfo, 'Boom 1', lineno=3)
    with pytest.raises(Exception) as excinfo:
        examples[1].evaluate()
    # check the line number of the second block:
    check_excinfo(examples[1], excinfo, 'Boom 2', lineno=9)
    examples[2].evaluate()
    # check the line number of the third block:
    assert namespace['foo'].__code__.co_firstlineno == 15
    return namespace['foo']


def test_no_future_imports():
    future_import_checks()


def test_single_future_import():
    future_import_checks('barry_as_FLUFL')


def test_multiple_future_imports():
    future_import_checks('barry_as_FLUFL', 'print_function')


def test_functional_future_imports():
    foo = future_import_checks('annotations')
    # This will keep working but not be an effective test once PEP 563 finally lands:
    assert foo.__code__.co_flags & __future__.annotations.compiler_flag


def test_blank_lines():
    examples, namespace = parse(
        'myst-codeblock-blank-lines.md', PythonCodeBlockParser(), expected=1
    )
    example, = examples
    compare(example.parsed, expected=''.join((
        '\n',
        'y = 0\n',
        '# now a blank line:\n',
        '\n',
        'y += 1\n',
        '# two blank lines:\n',
        '\n',
        '\n',
        "assert not y, 'Boom!'\n",
    )))
    with pytest.raises(AssertionError) as excinfo:
        example.evaluate()

    # check the line number in the exception:
    check_excinfo(example, excinfo, 'Boom!', lineno=12)


def test_blank_lines_indented():
    examples, namespace = parse(
        "myst-codeblock-blank-lines-indented.md", PythonCodeBlockParser(), expected=1
    )
    (example,) = examples
    compare(
        example.parsed,
        expected="".join(
            (
                "\n",
                "y = 0\n",
                "# now a blank line:\n",
                "\n",
                "y += 1\n",
                "# two blank lines:\n",
                "\n",
                "\n",
                "assert not y, 'Boom!'\n",
            )
        ),
    )
    with pytest.raises(AssertionError) as excinfo:
        example.evaluate()

    # check the line number in the exception:
    check_excinfo(example, excinfo, "Boom!", lineno=12)


def test_code_directive():
    sybil = Sybil([PythonCodeBlockParser()])
    check_path(sample_path('myst-code.md'), sybil, expected=1)


def test_sourcecode_directive():
    sybil = Sybil([PythonCodeBlockParser()])
    check_path(sample_path('myst-sourcecode.md'), sybil, expected=1)



from sybil.parsers.rest import ClearNamespaceParser, DocTestParser
from .helpers import parse


def test_basic():
    examples, namespace = parse('clear.txt', DocTestParser(), ClearNamespaceParser(), expected=4)
    for example in examples:
        example.evaluate()
    assert 'x' not in namespace, namespace



import importlib
import sys
from contextlib import contextmanager
from types import ModuleType
from typing import Iterator

from pathlib import Path


@contextmanager
def import_cleanup() -> Iterator[None]:
    """
    Clean up the results of importing modules, including the modification
    of :attr:`sys.path` necessary to do so.
    """
    modules = set(sys.modules)
    path = sys.path.copy()
    yield
    for added_module in set(sys.modules) - modules:
        sys.modules.pop(added_module)
    sys.path[:] = path
    importlib.invalidate_caches()


INIT_FILE = '__init__.py'


def import_path(path: Path) -> ModuleType:
    container = path
    while True:
        container = container.parent
        if not (container / INIT_FILE).exists():
            break
    relative = path.relative_to(container)
    if relative.name == INIT_FILE:
        parts = tuple(relative.parts)[:-1]
    else:
        parts = tuple(relative.parts)[:-1]+(relative.stem,)
    module = '.'.join(parts)
    try:
        return importlib.import_module(module)
    except ImportError as e:
        raise ImportError(
            f'{module!r} not importable from {path} as:\n{type(e).__name__}: {e}'
        ) from None



from typing import Sequence
import __future__

from sybil import Example


def pad(source: str, line: int) -> str:
    """
    Pad the supplied source such that line numbers will be based on the one provided
    when the source is evaluated.
    """
    # There must be a nicer way to get line numbers to be correct...
    return line  * '\n' + source


class PythonEvaluator:
    """
    The :any:`Evaluator` to use for :class:`Regions <sybil.Region>` containing
    Python source code.

    :param future_imports:
        A sequence of strings naming future import options, for example ``'annotations'``,
        that should be used at the top of each :class:`~sybil.Example` being evaluated.
    """

    def __init__(self, future_imports: Sequence[str] = ()) -> None:
        self.flags = 0
        for future_import in future_imports:
            self.flags |= getattr(__future__, future_import).compiler_flag

    def __call__(self, example: Example) -> None:
        # There must be a nicer way to get line numbers to be correct...
        source = pad(example.parsed, example.line + example.parsed.line_offset)
        code = compile(source, example.path, 'exec', flags=self.flags, dont_inherit=True)
        exec(code, example.namespace)
        # exec adds __builtins__, we don't want it:
        del example.namespace['__builtins__']



from typing import Iterable, Sequence, Optional, Callable

from sybil import Region, Document, Example
from sybil.typing import Evaluator, Lexer, Parser
from .doctest import DocTestStringParser
from .lexers import LexerCollection
from ...evaluators.doctest import DocTestEvaluator
from ...evaluators.python import PythonEvaluator


class AbstractCodeBlockParser:
    """
    An abstract parser for use when evaluating blocks of code.

    :param lexers:
        A sequence of :any:`Lexer` objects that will be applied in turn to each
        :class:`~sybil.Document`
        that is parsed. The :class:`~sybil.Region` objects returned by these lexers must have
        both an ``arguments`` string, containing the language of the lexed region, and a
        ``source`` :class:`~sybil.Lexeme` containing the source code of the lexed region.

    :param language:
        The language that this parser should look for. Lexed regions which don't have this
        language in their ``arguments`` lexeme will be ignored.

    :param evaluator:
        The evaluator to use for evaluating code blocks in the specified language.
        You can also override the :meth:`evaluate` method below.
    """

    language: str

    def __init__(
            self,
            lexers: Sequence[Lexer],
            language: Optional[str] = None,
            evaluator: Optional[Evaluator] = None,
    ) -> None:
        self.lexers = LexerCollection(lexers)
        if language is not None:
            self.language = language
        assert self.language, 'language must be specified!'
        self._evaluator: Optional[Evaluator] = evaluator

    def evaluate(self, example: Example) -> Optional[str]:
        """
        The :any:`Evaluator` used for regions yields by this parser can be provided by
        implementing this method.
        """
        raise NotImplementedError

    def __call__(self, document: Document) -> Iterable[Region]:
        for region in self.lexers(document):
            if region.lexemes['arguments'] == self.language:
                region.parsed = region.lexemes['source']
                region.evaluator = self._evaluator or self.evaluate
                yield region


class PythonDocTestOrCodeBlockParser:

    codeblock_parser_class: Callable[[str, Evaluator], Parser]

    def __init__(self, future_imports: Sequence[str] = (), doctest_optionflags: int = 0) -> None:
        self.doctest_parser = DocTestStringParser(
            DocTestEvaluator(doctest_optionflags)
        )
        self.codeblock_parser = self.codeblock_parser_class(
            'python', PythonEvaluator(future_imports)
        )

    def __call__(self, document: Document) -> Iterable[Region]:
        for region in self.codeblock_parser(document):
            source = region.parsed
            if region.parsed.startswith('>>>'):
                for doctest_region in self.doctest_parser(source, document.path):
                    doctest_region.adjust(region, source)
                    yield doctest_region
            else:
                yield region



import sys
from typing import Iterable
from unittest import SkipTest

import pytest
from testfixtures import ShouldRaise

from sybil import Example, Document, Region
from sybil.parsers.rest import PythonCodeBlockParser, DocTestParser, SkipParser
from .helpers import parse


def test_basic():
    examples, namespace = parse('skip.txt', PythonCodeBlockParser(), SkipParser(), expected=9)
    for example in examples:
        example.evaluate()
    assert namespace['run'] == [2, 5]


def test_conditional_edge_cases():
    examples, namespace = parse(
        'skip-conditional-edges.txt',
        DocTestParser(), SkipParser(),
        expected=8
    )
    namespace['sys'] = sys
    namespace['run'] = []
    skipped = []
    for example in examples:
        try:
            example.evaluate()
        except SkipTest as e:
            skipped.append(str(e))
    assert namespace['run'] == [1, 2, 3, 4]
    # we should always have one and only one skip from this document.
    assert skipped == ['skip 1']


def test_conditional_full():
    examples, namespace = parse('skip-conditional.txt', DocTestParser(), SkipParser(), expected=11)
    namespace['result'] = result = []
    for example in examples:
        try:
            example.evaluate()
        except SkipTest as e:
            result.append('skip:'+str(e))
    assert result == [
        'start',
        'skip:(2 > 1)',
        'good 1',
        'skip:foo',
        'skip:foo',
        'good 2',
        'skip:good reason',
    ]


def test_bad():
    examples, namespace = parse('skip-conditional-bad.txt', SkipParser(), expected=4)

    with pytest.raises(ValueError) as excinfo:
        examples[0].evaluate()
    assert str(excinfo.value) == 'Bad skip action: lolwut'

    examples[1].evaluate()

    with pytest.raises(ValueError) as excinfo:
        examples[2].evaluate()
    assert str(excinfo.value) == "Cannot have condition on 'skip: end'"

    with pytest.raises(SyntaxError):
        examples[3].evaluate()


def test_start_follows_start():
    examples, namespace = parse('skip-start-start.txt', DocTestParser(), SkipParser(), expected=7)
    namespace['result'] = result = []
    for example in examples[:2]:
        example.evaluate()
    with ShouldRaise(ValueError("'skip: start' cannot follow 'skip: start'")):
        examples[2].evaluate()
    assert result == []


def test_next_follows_start():
    examples, namespace = parse('skip-start-next.txt', DocTestParser(), SkipParser(), expected=7)
    namespace['result'] = result = []
    for example in examples[:2]:
        example.evaluate()
    with ShouldRaise(ValueError("'skip: next' cannot follow 'skip: start'")):
        examples[2].evaluate()
    assert result == []


def test_end_no_start():
    examples, namespace = parse('skip-just-end.txt', DocTestParser(), SkipParser(), expected=3)
    namespace['result'] = result = []
    examples[0].evaluate()
    with ShouldRaise(ValueError("'skip: end' must follow 'skip: start'")):
        examples[1].evaluate()
    assert result == ['good']


def test_next_follows_next():
    examples, namespace = parse('skip-next-next.txt', DocTestParser(), SkipParser(), expected=4)
    namespace['result'] = result = []
    for example in examples:
        example.evaluate()
    assert result == [1]



from sybil.parsers.myst import PythonCodeBlockParser, SkipParser
from .helpers import parse


def test_basic():
    examples, namespace = parse('myst-skip.md', PythonCodeBlockParser(), SkipParser(), expected=9)
    for example in examples:
        example.evaluate()
    assert namespace['run'] == [2, 5]



from .clear import AbstractClearNamespaceParser
from .codeblock import AbstractCodeBlockParser
from .skip import AbstractSkipParser
from .doctest import DocTestStringParser

__all__ = [
    'AbstractClearNamespaceParser',
    'AbstractCodeBlockParser',
    'AbstractSkipParser',
    'DocTestStringParser',
]



import sys
from os.path import join
from pathlib import Path
from unittest.main import main as unittest_main
from unittest.runner import TextTestRunner

import pytest
from pytest import main as pytest_main

from tests.helpers import add_to_python_path

DOC_EXAMPLES = Path(__file__).parent.parent / 'docs' / 'examples'


class CollectResults:
    def pytest_sessionfinish(self, session):
        self.session = session


def pytest_in(*path: str):
    results = CollectResults()
    return_code = pytest_main([join(DOC_EXAMPLES, *path)], plugins=[results])
    assert return_code == results.session.exitstatus
    return results.session


class TestIntegrationExamples:

    def test_pytest(self):
        session = pytest_in('integration', 'docs')
        assert session.exitstatus == 0
        assert session.testsfailed == 0
        assert session.testscollected == 3

    def test_unittest(self):
        runner = TextTestRunner(verbosity=2, stream=sys.stdout)
        path = str(DOC_EXAMPLES / 'integration' / 'unittest')
        main = unittest_main(
            exit=False, module=None, testRunner=runner,
            argv=['x', 'discover', '-s', path, '-t', path]
        )
        assert main.result.testsRun == 3
        assert len(main.result.failures) == 0
        assert len(main.result.errors) == 0


def test_quickstart():
    session = pytest_in('quickstart')
    assert session.exitstatus == 0
    assert session.testsfailed == 0
    assert session.testscollected == 4


def test_rest_text_rest_src():
    directory = 'rest_text_rest_src'
    with add_to_python_path(DOC_EXAMPLES / directory / 'src'):
        session = pytest_in(directory)
    assert session.testsfailed == 0
    assert session.testscollected == 5


def test_myst_text_rest_src():
    directory = 'myst_text_rest_src'
    with add_to_python_path(DOC_EXAMPLES / directory / 'src'):
        session = pytest_in(directory)
    assert session.testsfailed == 0
    assert session.testscollected == 5


def test_linting_and_checking():
    directory = 'linting_and_checking'
    with add_to_python_path(DOC_EXAMPLES / directory / 'src'):
        session = pytest_in(directory)
    assert session.testsfailed == 0
    assert session.testscollected == 3



from .clear import ClearNamespaceParser
from .codeblock import CodeBlockParser, PythonCodeBlockParser
from .skip import SkipParser

__all__ = [
    'ClearNamespaceParser',
    'CodeBlockParser',
    'PythonCodeBlockParser',
    'SkipParser',
]



from ..abstract import AbstractSkipParser
from .lexers import DirectiveInPercentCommentLexer
from ..markdown.lexers import DirectiveInHTMLCommentLexer


class SkipParser(AbstractSkipParser):
    """
    A :any:`Parser` for :ref:`skip <myst-skip-parser>` instructions.
    """

    def __init__(self) -> None:
        super().__init__([
            DirectiveInPercentCommentLexer('skip'),
            DirectiveInHTMLCommentLexer('skip'),
        ])



from sybil.evaluators.python import pad, PythonEvaluator
from sybil.parsers.abstract import AbstractCodeBlockParser
from sybil.parsers.rest.lexers import DirectiveLexer, DirectiveInCommentLexer
from sybil.typing import Evaluator
from typing import Optional, Sequence


class CodeBlockParser(AbstractCodeBlockParser):
    """
    A :any:`Parser` for :ref:`codeblock-parser` examples.

    :param language:
        The language that this parser should look for.

    :param evaluator:
        The evaluator to use for evaluating code blocks in the specified language.
        You can also override the :meth:`evaluate` method below.
    """

    def __init__(self, language: Optional[str] = None, evaluator: Optional[Evaluator] = None) -> None:
        super().__init__(
            [
                DirectiveLexer(directive=r'(sourcecode|code-block|code)'),
                DirectiveInCommentLexer(directive=r'(invisible-)?code(-block)?'),
            ],
            language, evaluator
        )

    pad = staticmethod(pad)


class PythonCodeBlockParser(CodeBlockParser):
    """
    A :any:`Parser` for Python :ref:`codeblock-parser` examples.

    :param future_imports:
        An optional sequence of strings that will be turned into
        ``from __future__ import ...`` statements and prepended to the code
        in each of the examples found by this parser.
    """

    def __init__(self, future_imports: Sequence[str] = ()) -> None:
        super().__init__(language='python', evaluator=PythonEvaluator(future_imports))



from testfixtures import compare

from sybil.parsers.markdown.lexers import RawFencedCodeBlockLexer
from sybil.region import Region
from .helpers import region_details, check_lexed_regions


def test_fenced_code_block():
    lexer = RawFencedCodeBlockLexer()
    check_lexed_regions('markdown-fenced-code-block.md', lexer, expected = [
        Region(12, 24, lexemes={'source': '<\n >\n'}),
        Region(34, 46, lexemes={'source': '<\n >\n'}),
        Region(177, 192, lexemes={'source': 'aaa\n~~~\n'}),
        Region(266, 285, lexemes={'source': 'aaa\n```\n'}),
        Region(301, 312, lexemes={'source': 'aaa\n'}),
        Region(296, 317, lexemes={'source': '~~~\naaa\n~~~\n'}),
        Region(397, 421, lexemes={'source': 'some stuff here\n~~~\n'}),
    ])



from __future__ import print_function

import re
from functools import partial
from os.path import split
from pathlib import Path

import pytest
from testfixtures import compare, StringComparison

from sybil import Sybil, Region
from sybil.document import Document, PythonDocument
from sybil.example import Example, SybilFailure

from .helpers import sample_path, write_doctest


@pytest.fixture()
def document():
    return Document('ABCDEFGH', '/the/path')


class TestRegion:

    def test_repr_with_evaluator(self):
        region = Region(0, 1, parsed='parsed', evaluator='evaluator')
        compare(
            repr(region),
            expected=(
                "<Region start=0 end=1 evaluator='evaluator'><Parsed>'parsed'</Parsed></Region>"
            )
        )

    def test_repr_with_lexemes(self):
        compare(
            str(Region(36, 56, lexemes={
                'language': 'python',
                'foo': None,
                'source': 'A'+'X'*1000+'Z',
                'bar': {},
                'baz': {f'a{i}': 'b' for i in range(11)},
            })),
            expected="<Region start=36 end=56>\n"
                     "language: 'python'\n"
                     "foo: None\n"
                     "source: 'AXXXXXXXXXXXXXXXXXXXX...XXXXXXXXXXXXXXXXXXXXZ'\n"
                     "bar: {}\n"
                     "baz: {'a0': 'b', 'a1': 'b', 'a2': 'b', 'a3': 'b', "
                     "'a4': 'b', 'a5': 'b', 'a6': 'b', 'a7': 'b', 'a8': 'b', 'a9': 'b', "
                     "'a10': 'b'}\n"
                     "</Region>"
        )

    def test_repr_with_parsed(self):
        compare(
            str(Region(36, 56, parsed={
                'language': 'python',
                'foo': None,
                'source': 'X'*1000,
                'bar': {},
                'baz': {f'a{i}': 'b' for i in range(11)},
            })),
            expected="<Region start=36 end=56><Parsed>{'language': 'python'"
                     "...9': 'b', 'a10': 'b'}}</Parsed></Region>"
        )


class TestExample:

    def test_repr(self, document):
        region = Region(0, 1, 'parsed', 'evaluator')
        example = Example(document, 1, 2, region, {})
        assert (repr(example) ==
                "<Example path=/the/path line=1 column=2 using 'evaluator'>")

    def test_evaluate_okay(self, document):
        def evaluator(example):
            example.namespace['parsed'] = example.parsed
        region = Region(0, 1, 'the data', evaluator)
        namespace = {}
        example = Example(document, 1, 2, region, namespace)
        result = example.evaluate()
        assert result is None
        assert namespace == {'parsed': 'the data'}

    def test_evaluate_not_okay(self, document):
        def evaluator(example):
            return 'foo!'
        region = Region(0, 1, 'the data', evaluator)
        example = Example(document, 1, 2, region, {})
        with pytest.raises(SybilFailure) as excinfo:
            example.evaluate()
        assert str(excinfo.value) == (
            'Example at /the/path, line 1, column 2 did not evaluate as '
            'expected:\nfoo!'
        )
        assert excinfo.value.example is example
        assert excinfo.value.result == 'foo!'

    def test_evaluate_raises_exception(self, document):
        def evaluator(example):
            raise ValueError('foo!')
        region = Region(0, 1, 'the data', evaluator)
        example = Example(document, 1, 2, region, {})
        with pytest.raises(ValueError) as excinfo:
            example.evaluate()
        assert str(excinfo.value) == 'foo!'


class TestDocument:

    def test_add(self, document):
        region = Region(0, 1, None, None)
        document.add(region)
        assert [e.region for e in document] == [region]

    def test_add_no_overlap(self, document):
        region1 = Region(0, 1, None, None)
        region2 = Region(6, 8, None, None)
        document.add(region1)
        document.add(region2)
        assert [e.region for e in document] == [region1, region2]

    def test_add_out_of_order(self, document):
        region1 = Region(0, 1, None, None)
        region2 = Region(6, 8, None, None)
        document.add(region2)
        document.add(region1)
        assert [e.region for e in document] == [region1, region2]

    def test_add_adjacent(self, document):
        region1 = Region(0, 1, None, None)
        region2 = Region(1, 2, None, None)
        region3 = Region(2, 3, None, None)
        document.add(region1)
        document.add(region3)
        document.add(region2)
        assert [e.region for e in document] == [region1, region2, region3]

    def test_add_before_start(self, document):
        region = Region(-1, 0, None, None)
        with pytest.raises(ValueError) as excinfo:
            document.add(region)
        assert str(excinfo.value) == (
            '<Region start=-1 end=0> '
            'from line 1, column 0 to line 1, column 1 '
            'is before start of document'
        )

    def test_add_after_end(self, document):
        region = Region(len(document.text), len(document.text)+1, None, None)
        with pytest.raises(ValueError) as excinfo:
            document.add(region)
        assert str(excinfo.value) == (
            '<Region start=8 end=9> '
            'from line 1, column 9 to line 1, column 10 '
            'goes beyond end of document'
        )

    def test_add_overlaps_with_previous(self, document):
        region1 = Region(0, 2, None, None)
        region2 = Region(1, 3, None, None)
        document.add(region1)
        with pytest.raises(ValueError) as excinfo:
            document.add(region2)
        assert str(excinfo.value) == (
            '<Region start=0 end=2>'
            ' from line 1, column 1 to line 1, column 3 overlaps '
            '<Region start=1 end=3>'
            ' from line 1, column 2 to line 1, column 4'
        )

    def test_add_at_same_place(self, document):
        region1 = Region(0, 2, None, None)
        region2 = Region(0, 3, None, None)
        document.add(region1)
        with pytest.raises(ValueError) as excinfo:
            document.add(region2)
        assert str(excinfo.value) == (
            '<Region start=0 end=3>'
            ' from line 1, column 1 to line 1, column 4 overlaps '
            '<Region start=0 end=2>'
            ' from line 1, column 1 to line 1, column 3'
        )

    def test_add_identical(self, document):
        region1 = Region(0, 2, None, None)
        region2 = Region(0, 2, None, None)
        document.add(region1)
        with pytest.raises(ValueError) as excinfo:
            document.add(region2)
        assert str(excinfo.value) == (
            '<Region start=0 end=2>'
            ' from line 1, column 1 to line 1, column 3 overlaps '
            '<Region start=0 end=2>'
            ' from line 1, column 1 to line 1, column 3'
        )

    def test_add_overlaps_with_next(self, document):
        region1 = Region(0, 1, None, None)
        region2 = Region(1, 3, None, None)
        region3 = Region(2, 4, None, None)
        document.add(region1)
        document.add(region3)
        with pytest.raises(ValueError) as excinfo:
            document.add(region2)
        assert str(excinfo.value) == (
            '<Region start=1 end=3> '
            'from line 1, column 2 to line 1, column 4 overlaps '
            '<Region start=2 end=4> '
            'from line 1, column 3 to line 1, column 5'
        )

    def test_example_path(self, document):
        document.add(Region(0, 1, None, None))
        assert [e.document for e in document] == [document]

    def test_example_line_and_column(self):
        text = 'R1XYZ\nR2XYZ\nR3XYZ\nR4XYZ\nR4XYZ\n'
        i = text.index
        document = Document(text, '')
        document.add(Region(0,         i('R2')+2, None, None))
        document.add(Region(i('R3')-1, i('R3')+2, None, None))
        document.add(Region(i('R4')+3, len(text), None, None))
        assert ([(e.line, e.column) for e in document] ==
                [(1, 1), (2, 6), (4, 4)])


def check(letter, parsed, namespace):
    assert namespace is None
    text, expected = parsed
    assert set(text) == {letter}
    actual = text.count(letter)
    if actual != expected:
        return '{} count was {} instead of {}'.format(
            letter, actual, expected
        )
    # This would normally be wrong, but handy for testing here:
    return '{} count was {}, as expected'.format(letter, actual)


def parse_for_x(document):
    for m in re.finditer(r'(X+) (\d+) check', document.text):
        yield Region(m.start(), m.end(),
                     (m.group(1), int(m.group(2))),
                     partial(check, 'X'))


def parse_for_y(document):
    for m in re.finditer(r'(Y+) (\d+) check', document.text):
        yield Region(m.start(), m.end(),
                     (m.group(1), int(m.group(2))),
                     partial(check, 'Y'))


def evaluate_examples(examples):
    return [e.region.evaluator(e.region.parsed, namespace=None)
            for e in examples]


class TestSybil:

    def test_parse(self):
        sybil = Sybil([parse_for_x, parse_for_y])
        document = sybil.parse(Path(sample_path('sample1.txt')))
        assert (evaluate_examples(document) ==
                ['X count was 4, as expected',
                 'Y count was 3, as expected'])
        document = sybil.parse(Path(sample_path('sample2.txt')))
        assert (evaluate_examples(document) ==
                ['X count was 3 instead of 4',
                 'Y count was 3, as expected'])

    def test_explicit_encoding(self, tmp_path: Path):
        path = (tmp_path / 'encoded.txt')
        path.write_text(u'X 1 check\n\xa3', encoding='charmap')
        sybil = Sybil([parse_for_x], encoding='charmap')
        document = sybil.parse(path)
        assert (evaluate_examples(document) ==
                ['X count was 1, as expected'])

    def test_augment_document_mapping(self, tmp_path: Path):

        class TextDocument(Document):
            pass

        sybil = Sybil([], document_types={'.txt': TextDocument})
        document = sybil.parse(write_doctest(tmp_path, 'test.txt'))
        assert type(document) is TextDocument
        document = sybil.parse(write_doctest(tmp_path, 'test.rst'))
        assert type(document) is Document

    def test_override_document_mapping(self):
        sybil = Sybil([parse_for_x, parse_for_y], document_types={'.py': PythonDocument})
        document = sybil.parse(Path(sample_path('sample1.txt')))
        assert (evaluate_examples(document) ==
                ['X count was 4, as expected',
                 'Y count was 3, as expected'])
        document = sybil.parse(Path(sample_path('comments.py')))
        assert (evaluate_examples(document) ==
                ['X count was 4, as expected',
                 'Y count was 3, as expected'])

    def test_addition(self):
        rest = Sybil([parse_for_x])
        myst = Sybil([parse_for_y])
        sybil = rest + myst
        assert sybil == [rest, myst]
        # check integrations exist:
        assert sybil.pytest
        assert sybil.unittest

    def test_addition_to_collection(self):
        rest = Sybil([parse_for_x])
        myst = Sybil([parse_for_y])
        bust = Sybil([parse_for_y])
        sybil = rest + myst
        sybil += [bust]
        assert sybil == [rest, myst, bust]
        # check integrations exist:
        assert sybil.pytest
        assert sybil.unittest

    def test_default_str(self):
        sybil = Sybil([parse_for_x])
        compare(str(sybil), expected=StringComparison('<Sybil: \d+>'))

    def test_named_str(self):
        sybil = Sybil([parse_for_x], name='foo')
        compare(str(sybil), expected='<Sybil: foo>')

    def test_default_repr(self):
        sybil = Sybil([parse_for_x])
        compare(repr(sybil), expected=StringComparison('<Sybil: \d+>'))

    def test_named_repr(self):
        sybil = Sybil([parse_for_x], name='foo')
        compare(repr(sybil), expected='<Sybil: foo>')


def check_into_namespace(example):
    parsed, namespace = example.region.parsed, example.namespace
    if 'parsed' not in namespace:
        namespace['parsed'] = []
    namespace['parsed'].append(parsed)
    print(namespace['parsed'])


def parse(document):
    for m in re.finditer(r'([XY]+) (\d+) check', document.text):
        yield Region(m.start(), m.end(), m.start(), check_into_namespace)


def test_namespace(capsys):
    sybil = Sybil([parse], path='./samples')
    documents = [sybil.parse(p) for p in sorted(sybil.path.glob('sample*.txt'))]
    actual = []
    for document in documents:
        for example in document:
            print(split(example.path)[-1], example.line)
            example.evaluate()
            actual.append((
                split(example.path)[-1],
                example.line,
                document.namespace['parsed'].copy(),
            ))
    out, _ = capsys.readouterr()
    assert out.split('\n') == [
        'sample1.txt 1',
        '[0]',
        'sample1.txt 3',
        '[0, 14]',
        'sample2.txt 1',
        '[0]',
        'sample2.txt 3',
        '[0, 13]',
        ''
    ]



# See docs/license.rst for license details.
# Copyright (c) 2017 onwards Chris Withers

import os

from setuptools import setup, find_packages

base_dir = os.path.dirname(__file__)

setup(
    name='sybil',
    version='7.0.0',
    author='Chris Withers',
    author_email='chris@withers.org',
    license='MIT',
    description="Automated testing for the examples in your code and documentation.",
    long_description=open('README.rst').read(),
    url='https://github.com/simplistix/sybil',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    packages=find_packages(exclude=['tests', 'functional_tests']),
    package_data={"sybil": ["py.typed"]},
    python_requires=">=3.8",
    extras_require=dict(
        test=[
            'mypy',
            'myst_parser',
            'pytest>=7.1.0',
            'pytest-cov',
            'seedir',
            'testfixtures',
            'types-PyYAML',
            ],
        build=[
            'furo',
            'sphinx',
            'twine',
            'urllib3<2',  # workaround for RTD builds failing with old SSL
            'wheel',
        ]
    ),
)



# Tests for backwards compatibility
import json

from sybil.parsers.capture import parse_captures
from sybil.parsers.codeblock import CodeBlockParser, PythonCodeBlockParser
from sybil.parsers.skip import skip
from .helpers import parse


def test_imports():
    # uncomment once all the moves are done!
    from sybil.parsers.capture import parse_captures
    from sybil.parsers.codeblock import CodeBlockParser, PythonCodeBlockParser
    from sybil.parsers.doctest import DocTestParser
    from sybil.parsers.skip import skip
    pass


def test_code_block_parser_pad():
    assert CodeBlockParser('foo').pad('x', line=2) == '\n\nx'


def test_skip_parser_function():
    examples, namespace = parse('skip.txt', PythonCodeBlockParser(), skip, expected=9)
    for example in examples:
        example.evaluate()
    assert namespace['run'] == [2, 5]


def test_capture_parser_function():
    examples, namespace = parse('capture_codeblock.txt', parse_captures, expected=1)
    examples[0].evaluate()
    assert json.loads(namespace['json']) == {"a key": "value", "b key": 42}



import re
from typing import Iterable, Sequence

from sybil import Document, Region
from sybil.evaluators.skip import Skipper
from sybil.parsers.abstract.lexers import LexerCollection
from sybil.typing import Lexer

SKIP_ARGUMENTS_PATTERN = re.compile(r'(\w+)(?:\s+(.+)$)?')


class AbstractSkipParser:
    """
    An abstract parser for skipping subsequent examples.

    :param lexers:
        A sequence of :any:`Lexer` objects that will be applied in turn to each
        :class:`~sybil.Document` that is parsed.
    """

    def __init__(self, lexers: Sequence[Lexer]):
        self.lexers = LexerCollection(lexers)
        self.skipper = Skipper()

    def __call__(self, document: Document) -> Iterable[Region]:
        for lexed in self.lexers(document):
            match = SKIP_ARGUMENTS_PATTERN.match(lexed.lexemes['arguments'])
            yield Region(lexed.start, lexed.end, match.groups(), self.skipper)



from testfixtures import compare

from sybil.parsers.rest.lexers import DirectiveInCommentLexer
from sybil.parsers.rest.lexers import DirectiveLexer
from sybil.region import Region
from .helpers import lex, lex_text


def test_examples_from_parsing_tests():
    lexer = DirectiveLexer(directive='code-block', arguments='python')
    compare(lex('codeblock.txt', lexer)[:2], expected=[
        Region(23, 56, lexemes={
            'directive': 'code-block',
            'arguments': 'python',
            'options': {},
            'source': 'y += 1\n',
        }),
        Region(106, 157, lexemes={
            'directive': 'code-block',
            'arguments': 'python',
            'options': {},
            'source': "raise Exception('boom!')\n",
        }),
    ])


def test_examples_from_directive_tests():
    lexer = DirectiveLexer(directive='doctest')
    compare(lex('doctest_directive.txt', lexer), expected=[
        Region(102, 136, lexemes={
            'directive': 'doctest',
            'arguments': None,
            'options': {},
            'source': '>>> 1 + 1\n2\n',
        }),
        Region(205, 249, lexemes={
            'directive': 'doctest',
            'arguments': None,
            'options': {},
            'source': '>>> 1 + 1\nUnexpected!\n',
        }),
        Region(307, 353, lexemes={
            'directive': 'doctest',
            'arguments': None,
            'options': {},
            'source': ">>> raise Exception('boom!')",
        }),
    ])


def test_directive_nested_in_md():
    lexer = DirectiveLexer(directive='doctest')
    compare(lex('doctest_rest_nested_in_md.md', lexer), expected=[
        Region(14, 47, lexemes={
            'directive': 'doctest',
            'arguments': None,
            'options': {},
            'source': '>>> 1 + 1\n3',
        }),
    ])


def test_directive_with_single_line_body_at_end_of_string():
    text = """
        My comment
    
        .. code-block:: python
            test_my_code("Hello World")
    """
    lexer = DirectiveLexer(directive='code-block')
    start = 25
    end = 95
    compare(lex_text(text, lexer), expected=[
        Region(start, end, lexemes={
            'directive': 'code-block',
            'arguments': 'python',
            'options': {},
            'source': 'test_my_code("Hello World")',
        }),
    ])
    compare(text[start:end], show_whitespace=True, expected=(
        '        .. code-block:: python\n'
        '            test_my_code("Hello World")'
    ))


def test_directive_with_multi_line_body_at_end_of_string():
    text = """
        My comment

        .. code-block:: python
            test_my_code("Hello")
            test_my_code("World")
    """
    lexer = DirectiveLexer(directive='code-block')
    start = 21
    end = 119
    compare(lex_text(text, lexer), expected=[
        Region(start, end, lexemes={
            'directive': 'code-block',
            'arguments': 'python',
            'options': {},
            'source': 'test_my_code("Hello")\ntest_my_code("World")',
        }),
    ])
    compare(text[start:end], show_whitespace=True, expected=(
        '        .. code-block:: python\n'
        '            test_my_code("Hello")\n'
        '            test_my_code("World")'
    ))


def test_skip_lexing():
    lexer = DirectiveInCommentLexer(directive='skip')
    compare(lex('skip.txt', lexer), expected=[
        Region(70, 84, lexemes={
            'directive': 'skip', 'arguments': 'next', 'options': {}, 'source': ''
        }),
        Region(197, 212, lexemes={
            'directive': 'skip', 'arguments': 'start', 'options': {}, 'source': ''
        }),
        Region(329, 342, lexemes={
            'directive': 'skip', 'arguments': 'end', 'options': {}, 'source': ''
        }),
    ])


def test_skip_lexing_bad():
    lexer = DirectiveInCommentLexer(directive='skip')
    compare(lex('skip-conditional-bad.txt', lexer), expected=[
        Region(13, 29, lexemes={
            'directive': 'skip', 'arguments': 'lolwut', 'options': {}, 'source': ''
        }),
        Region(49, 64, lexemes={
            'directive': 'skip', 'arguments': 'start', 'options': {}, 'source': ''
        }),
        Region(65, 88, lexemes={
            'directive': 'skip', 'arguments': 'end if(1 > 2)', 'options': {}, 'source': ''
        }),
        Region(104, 179, lexemes={
            'directive': 'skip',
            'arguments': 'next if(sys.version_info < (3, 0), reason="only true on python 3"',
            'options': {},
            'source': '',
        }),
    ])


def test_arguments_without_body():
    lexer = DirectiveInCommentLexer(directive='skip')
    compare(lex('skip-conditional-edges.txt', lexer), expected=[
        Region(0, 40, lexemes={
            'directive': 'skip',
            'arguments': 'next if(True, reason="skip 1")',
            'options': {},
            'source': ''
        }),
        Region(100, 142, lexemes={
            'directive': 'skip',
            'arguments': 'start if(False, reason="skip 2")',
            'options': {},
            'source': ''
        }),
        Region(205, 218, lexemes={
            'directive': 'skip',
            'arguments': 'end',
            'options': {},
            'source': ''
        }),
    ])


def test_lexing_directives():
    lexer = DirectiveLexer('[^:]+')
    compare(lex('lexing-directives.txt', lexer), expected=[
        Region(55, 245, lexemes={
            'directive': 'note',
            'arguments': 'This is a note admonition.',
            'options': {},
            'source': ('This is the second line of the first paragraph.\n'
                       '\n'
                       '- The note contains all indented body elements\n'
                       '  following.\n'
                       '- It includes this bullet list.\n'),
        }),
        Region(246, 326, lexemes={
            'directive': 'admonition',
            'arguments': 'And, by the way...',
            'options': {},
            'source': 'You can make up your own admonition too.\n',
        }),
        Region(327, 389, lexemes={
            'directive': 'sample',
            'arguments': None,
            'options': {},
            'source': 'This directive has no arguments, just a body.\n',
        }),
        Region(457, 480, lexemes={
            'directive': 'image',
            'arguments': 'picture.png',
            'options': {},
            'source': '',
        }),
        Region(481, 598, lexemes={
            'directive': 'image',
            'arguments': 'picture.jpeg',
            'options': {
                'height': '100px',
                'width': '200 px',
                'scale': '50 %',
                'alt': 'alternate text',
                'align': 'right',
            },
            'source': '',
        }),
        Region(599, 1353, lexemes={
            'directive': 'figure',
            'arguments': 'picture.png',
            'options': {
                'alt': 'map to buried treasure',
                'scale': '50 %',
            },
            'source': ('This is the caption of the figure (a simple paragraph).\n'
                       '\n'
                       'The legend consists of all elements after the caption.  In this\n'
                       'case, the legend consists of this paragraph and the following\n'
                       'table:\n'
                       '\n'
                       '+-----------------------+-----------------------+\n'
                       '| Symbol                | Meaning               |\n'
                       '+=======================+=======================+\n'
                       '| .. image:: tent.png   | Campground            |\n'
                       '+-----------------------+-----------------------+\n'
                       '| .. image:: waves.png  | Lake                  |\n'
                       '+-----------------------+-----------------------+\n'
                       '| .. image:: peak.png   | Mountain              |\n'
                       '+-----------------------+-----------------------+\n'
                       '\n')
        }),
        Region(1354, 1486, lexemes={
            'directive': 'topic',
            'arguments': 'Topic Title',
            'options': {},
            'source': ('Subsequent indented lines comprise\n'
                       'the body of the topic, and are\n'
                       'interpreted as body elements.\n')
        }),
        Region(1539, 1616, lexemes={
            'directive': 'topic',
            'arguments': 'example.cfg',
            'options': {'class': 'read-file'},
            'source': '::\n\n  [A Section]\n  dir = frob\n'
        }),
        Region(1635, 1819, lexemes={
            'directive': 'sidebar',
            'arguments': 'Optional Sidebar Title',
            'options': {'subtitle': 'Optional Sidebar Subtitle'},
            'source': ('Subsequent indented lines comprise\n'
                       'the body of the sidebar, and are\n'
                       'interpreted as body elements.\n')
        }),
        Region(1856, 1871, lexemes={
            'directive': 'skip',
            'arguments': 'next',
            'options': {},
            'source': '',
        }),
        Region(1871, 1911, lexemes={
            'directive': 'code-block',
            'arguments': 'python',
            'options': {},
            'source': (
                'run.append(1)\n'
            ),
        }),
        Region(1988, 2066, lexemes={
            'directive': 'topic',
            'arguments': 'example.cfg',
            'options': {'class': 'read-file'},
            'source': (
                '::\n\n  [A Section]\n  dir = frob\n\n'
            ),
        }),
    ])


def test_lexing_nested_directives():
    # This test illustrates that we can lex nested directives,
    # but the blocks overlap, so would not be parseable unless only
    # the deepest nested blocks were parsed.
    lexer = DirectiveLexer('[^:]+')
    actual = lex('lexing-nested-directives.txt', lexer)
    compare(actual, expected=[
        Region(55, 236, lexemes={
            'directive': 'note',
            'arguments': 'This is a note admonition.',
            'options': {},
            'source': (
                'This is the second line of the first paragraph.\n\n'
                '   .. admonition:: And, by the way...\n\n'
                '      You can make up your own admonition too.\n'
            ),
        }),
        Region(144, 236, lexemes={
            'directive': 'admonition',
            'arguments': 'And, by the way...',
            'options': {},
            'source': 'You can make up your own admonition too.\n',
        }),
        Region(304, 783, lexemes={
            'directive': 'image',
            'arguments': 'picture.png',
            'options': {},
            'source': (
                '.. image:: picture.jpeg\n'
                '   :height: 100px\n'
                '   :width: 200 px\n'
                '   :scale: 50 %\n'
                '   :alt: alternate text\n'
                '   :align: right\n'
                '\n'
                '.. figure:: picture.png\n'
                '   :scale: 50 %\n'
                '   :alt: map to buried treasure\n'
                '\n'
                '   This is the caption of the figure (a simple paragraph).\n'
                '\n'
                '\n'
                '    .. topic:: Topic Title\n'
                '\n'
                '        Subsequent indented lines comprise\n'
                '        the body of the topic, and are\n'
                '        interpreted as body elements.'
            ),
        }),
        Region(328, 469, lexemes={
            'directive': 'image',
            'arguments': 'picture.jpeg',
            'options': {
                'align': 'right',
                'alt': 'alternate text',
                'height': '100px',
                'scale': '50 %',
                'width': '200 px',
            },
            'source': '',
        }),
        Region(470, 783, lexemes={
            'directive': 'figure',
            'arguments': 'picture.png',
            'options': {
                'scale': '50 %',
                'alt': 'map to buried treasure'
            },
            'source': (
                'This is the caption of the figure (a simple paragraph).\n'
                '\n'
                '\n'
                ' .. topic:: Topic Title\n'
                '\n'
                '     Subsequent indented lines comprise\n'
                '     the body of the topic, and are\n'
                '     interpreted as body elements.'
            ),
        }),
        Region(620, 783, lexemes={
            'directive': 'topic',
            'arguments': 'Topic Title',
            'options': {},
            'source': (
                'Subsequent indented lines comprise\n'
                'the body of the topic, and are\n'
                'interpreted as body elements.'
            ),
        }),
    ])


def test_lexing_directive_in_comment_without_double_colon():
    lexer = DirectiveInCommentLexer('clear-namespace')
    compare(lex('clear.txt', lexer), expected=[
        Region(48, 67, lexemes={
            'directive': 'clear-namespace', 'arguments': None, 'options': {}, 'source': ''
        }),
    ])



"""
This is a module doc string.
"""  # This is a comment on how annoying 3.7 and earlier are!
import re
from functools import partial
from pathlib import Path
from typing import Iterable

from testfixtures import compare, ShouldRaise

from sybil import Region, Example
from sybil.document import PythonDocStringDocument, Document
from sybil.example import NotEvaluated, SybilFailure
from sybil.exceptions import LexingException
from .helpers import ast_docstrings, parse, sample_path


def test_extract_docstring():
    """
    This is a function docstring.
    """
    python_source_code = Path(__file__).read_text()
    expected = list(ast_docstrings(python_source_code))
    actual = list(PythonDocStringDocument.extract_docstrings(python_source_code))
    compare(expected, actual=[text for _, _, text in actual], show_whitespace=True)
    compare(expected, actual=[python_source_code[s:e] for s, e, _ in actual], show_whitespace=True)


def test_all_docstrings_extracted_correctly(python_file):
    problems = []
    path, source = python_file
    expected = list(ast_docstrings(source))
    actual = list(PythonDocStringDocument.extract_docstrings(source))
    actual_from_start_end = [source[s:e].replace('\\"', '"') for s, e, _ in actual]
    check = partial(compare, raises=False, show_whitespace=True)
    for check_result in (
        check(expected=len(expected), actual=len(actual), prefix=f'{path}: number'),
        check(expected, actual=[text for _, _, text in actual], prefix=f'{path}: content'),
        check(expected, actual=actual_from_start_end, prefix=f'{path}: start/end'),
    ):
        if check_result:  # pragma: no cover - Only on failures!
            problems.append(check_result)
    if problems:  # pragma: no cover - Only on failures!
        raise AssertionError('docstrings not correctly extracted:\n\n'+'\n\n'.join(problems))


def test_evaluator_returns_non_string():

    def evaluator(example: Example) -> NotEvaluated:
        # This is a bug!
        return NotEvaluated()

    def parser(document: Document) -> Iterable[Region]:
        yield Region(0, 1, None, evaluator)

    examples, namespace = parse('sample1.txt', parser, expected=1)
    with ShouldRaise(SybilFailure(examples[0], f'NotEvaluated()')):
        examples[0].evaluate()


def test_nested_evaluators():

    def record(example: Example, evaluator_name):
        (instruction, id_, param) = example.parsed
        example.namespace['results'].append((instruction, id_, param, evaluator_name))

    def normal_evaluator(example: Example):
        record(example, 'normal')

    class InstructionEvaluator:
        def __init__(self, id_, mode):
            self.name = f'{id_}-{mode}'
            self.mode = mode
            assert hasattr(self, mode), mode

        def __call__(self, example: Example):
            (instruction, id_, param) = example.parsed
            if instruction in ('install', 'remove'):
                raise NotEvaluated()
            record(example, self.name)
            return getattr(self, self.mode)(id_)

        def install(self, example: Example):
            record(example, self.name+'-install')
            example.document.push_evaluator(self)

        def remove(self, example: Example):
            record(example, self.name + '-remove')
            example.document.pop_evaluator(self)

        def passthrough(self, id_: int):
            raise NotEvaluated()

        def all(self, id_: int):
            return ''

        def even(self, id_: int):
            if id_ % 2:
                raise NotEvaluated()

    evaluators = {}

    def parser(document: Document) -> Iterable[Region]:
        for match in re.finditer(r'(example|install|remove)-(\d+)(.+)?', document.text):
            instruction, id_, param = match.groups()
            id_ = int(id_)
            param = param.lstrip('-') if param else None
            if instruction == 'example':
                evaluator = normal_evaluator
            else:
                obj = evaluators.get(id_)
                if obj is None:
                    obj = evaluators.setdefault(id_, InstructionEvaluator(id_, param))
                evaluator = getattr(obj, instruction)
            yield Region(match.start(), match.end(), (instruction, id_, param), evaluator)

    examples, namespace = parse('nested-evaluators.txt', parser, expected=12)
    namespace['results'] = results = []
    for e in examples:
        e.evaluate()
    compare(results, expected=[
        ('example', 0, None, 'normal'),
        ('install', 1, 'passthrough', '1-passthrough-install'),
        ('example', 2, None, '1-passthrough'),
        ('example', 2, None, 'normal'),
        ('install', 3, 'all', '3-all-install'),
        ('example', 4, None, '3-all'),
        ('install', 5, 'even', '5-even-install'),
        ('example', 6, None, '5-even'),
        ('example', 7, None, '5-even'),
        ('example', 7, None, '3-all'),
        ('remove', 5, None, '5-even-remove'),
        ('example', 8, None, '3-all'),
        ('remove', 3, None, '3-all-remove'),
        ('example', 9, None, '1-passthrough'),
        ('example', 9, None, 'normal'),
    ])


def test_nested_evaluators_not_evaluated_from_region():

    def evaluator(example: Example):
        raise NotEvaluated()

    def parser(document: Document) -> Iterable[Region]:
        yield Region(0, 1, None, evaluator)

    examples, namespace = parse('sample1.txt', parser, expected=1)
    with ShouldRaise(SybilFailure(examples[0], f'{evaluator!r} should not raise NotEvaluated()')):
        examples[0].evaluate()


def test_find_region_sources_bad_end_match():
    path = sample_path('lexing-fail.txt')
    document = Document(Path(path).read_text(), path)
    with ShouldRaise(LexingException(f"Could not match 'END' in {path}:\n'\\nEDN\\n'")):
        tuple(document.find_region_sources(
            start_pattern=re.compile('START'), end_pattern=re.compile('END'),
        ))



from doctest import (
    DocTestParser as BaseDocTestParser,
    Example as DocTestExample,
)
from typing import Iterable

from sybil.evaluators.doctest import DocTestEvaluator
from sybil.region import Region


class DocTestStringParser(BaseDocTestParser):
    """
    This isn't a true :any:`Parser` in that it must be called with a :class:`str` containing
    the doctest example's source and the file name that the example came from.
    """

    def __init__(self, evaluator: DocTestEvaluator = DocTestEvaluator()) -> None:
        #: The evaluator to use for any doctests found in the supplied source string.
        self.evaluator: DocTestEvaluator = evaluator

    def __call__(self, string: str, name: str) -> Iterable[Region]:
        """
        This will yield :class:`sybil.Region` objects for any doctest examples found in
        the supplied ``string`` with the :attr:`evaluator` supplied to its constructor
        and the file ``name`` supplied.

        Each section starting with a ``>>>`` will form a separate region.
        """
        # a cut down version of doctest.DocTestParser.parse:
        charno, lineno = 0, 0
        # Find all doctest examples in the string:
        for m in self._EXAMPLE_RE.finditer(string):  # type: ignore
            # Update lineno (lines before this example)
            lineno += string.count('\n', charno, m.start())
            # Extract info from the regexp match.
            source, options, want, exc_msg = self._parse_example(m, name, lineno)  # type: ignore

            # Create an Example, and add it to the list.
            if not self._IS_BLANK_OR_COMMENT(source):  # type: ignore
                yield Region(
                    m.start(),
                    m.end(),
                    DocTestExample(source, want, exc_msg,
                                   lineno=lineno,
                                   indent=len(m.group('indent')),
                                   options=options),
                    self.evaluator

                )
            # Update lineno (lines inside this example)
            lineno += string.count('\n', m.start(), m.end())
            # Update charno.
            charno = m.end()





import __future__
import sys
from pathlib import Path

import pytest
from testfixtures import compare

from sybil import Example, Sybil, Region
from sybil.document import Document
from sybil.parsers.codeblock import PythonCodeBlockParser, CodeBlockParser
from sybil.parsers.rest import DocTestParser
from .helpers import check_excinfo, parse, sample_path, check_path, SAMPLE_PATH, add_to_python_path


def test_basic():
    examples, namespace = parse('codeblock.txt', PythonCodeBlockParser(), expected=7)
    namespace['y'] = namespace['z'] = 0
    assert examples[0].evaluate() is None
    assert namespace['y'] == 1
    assert namespace['z'] == 0
    with pytest.raises(Exception) as excinfo:
        examples[1].evaluate()
    compare(examples[1].parsed, expected="raise Exception('boom!')\n", show_whitespace=True)
    assert examples[1].line == 9
    check_excinfo(examples[1], excinfo, 'boom!', lineno=11)
    assert examples[2].evaluate() is None
    assert namespace['y'] == 1
    assert namespace['z'] == 1
    assert examples[3].evaluate() is None
    assert namespace['bin'] == b'x'
    assert namespace['uni'] == u'x'
    assert examples[4].evaluate() is None
    assert 'NoVars' in namespace
    assert examples[5].evaluate() is None
    assert namespace['define_this'] == 1
    assert examples[6].evaluate() is None
    assert 'YesVars' in namespace
    assert '__builtins__' not in namespace


def test_other_language_composition_pass():

    def oh_hai(example):
        assert isinstance(example, Example)
        assert 'HAI' in example.parsed

    parser = CodeBlockParser(language='lolcode', evaluator=oh_hai)
    examples, namespace = parse('codeblock.txt', parser, expected=1)
    assert examples[0].evaluate() is None


def test_other_language_composition_fail():
    def oh_noez(example):
        if 'KTHXBYE' in example.parsed:
            raise ValueError('oh noez')

    parser = CodeBlockParser(language='lolcode', evaluator=oh_noez)
    examples, namespace = parse('codeblock.txt', parser, expected=1)
    with pytest.raises(ValueError):
        examples[0].evaluate()


def test_other_language_no_evaluator():
    parser = CodeBlockParser('foo')
    with pytest.raises(NotImplementedError):
        parser.evaluate(...)


class LolCodeCodeBlockParser(CodeBlockParser):

    language = 'lolcode'

    def evaluate(self, example: Example):
        if example.parsed != 'HAI\n':
            raise ValueError(repr(example.parsed))


def test_other_language_inheritance():
    examples, namespace = parse('codeblock_lolcode.txt', LolCodeCodeBlockParser(), expected=2)
    examples[0].evaluate()
    with pytest.raises(ValueError) as excinfo:
        examples[1].evaluate()
    assert str(excinfo.value) == "'KTHXBYE'"


class IgnoringPythonCodeBlockParser(PythonCodeBlockParser):

    def __call__(self, document):
        for region in super().__call__(document):
            options = region.lexemes.get('options')
            if options and 'ignore' in options:
                region.evaluator = None
            yield region


def test_other_functionality_inheritance():
    examples, namespace = parse(
        'codeblock-subclassing.txt', IgnoringPythonCodeBlockParser(), expected=2
    )
    examples[0].evaluate()
    examples[1].evaluate()


def future_import_checks(*future_imports):
    parser = PythonCodeBlockParser(future_imports)
    examples, namespace = parse('codeblock_future_imports.txt', parser, expected=3)
    with pytest.raises(Exception) as excinfo:
        examples[0].evaluate()
    # check the line number of the first block, which is the hardest case:
    check_excinfo(examples[0], excinfo, 'Boom 1', lineno=3)
    with pytest.raises(Exception) as excinfo:
        examples[1].evaluate()
    # check the line number of the second block:
    check_excinfo(examples[1], excinfo, 'Boom 2', lineno=9)
    examples[2].evaluate()
    # check the line number of the third block:
    assert namespace['foo'].__code__.co_firstlineno == 15
    return namespace['foo']


def test_no_future_imports():
    future_import_checks()


def test_single_future_import():
    future_import_checks('barry_as_FLUFL')


def test_multiple_future_imports():
    future_import_checks('barry_as_FLUFL', 'print_function')


def test_functional_future_imports():
    foo = future_import_checks('annotations')
    # This will keep working but not be an effective test once PEP 563 finally lands:
    assert foo.__code__.co_flags & __future__.annotations.compiler_flag


def test_windows_line_endings(tmp_path: Path):
    p = tmp_path / "example.txt"
    p.write_bytes(
        b'This is my example:\r\n\r\n'
        b'.. code-block:: python\r\n\r\n'
        b'    from math import cos\r\n'
        b'    x = 123\r\n\r\n'
        b'That was my example.\r\n'
    )
    document = Document.parse(str(p), PythonCodeBlockParser())
    example, = document
    example.evaluate()
    assert document.namespace['x'] == 123


def test_line_numbers_with_options():
    parser = PythonCodeBlockParser()
    examples, namespace = parse('codeblock_with_options.txt', parser, expected=2)
    with pytest.raises(Exception) as excinfo:
        examples[0].evaluate()
    # check the line number of the first block, which is the hardest case:
    check_excinfo(examples[0], excinfo, 'Boom 1', lineno=6)
    with pytest.raises(Exception) as excinfo:
        examples[1].evaluate()
    # check the line number of the second block:
    check_excinfo(examples[1], excinfo, 'Boom 2', lineno=14)


def test_codeblocks_in_docstrings():
    sybil = Sybil([PythonCodeBlockParser()])
    with add_to_python_path(SAMPLE_PATH):
        check_path(sample_path('docstrings.py'), sybil, expected=3)


def test_doctest_in_docstrings():
    sybil = Sybil([DocTestParser()])
    with add_to_python_path(SAMPLE_PATH):
        check_path(sample_path('docstrings.py'), sybil, expected=1)


def test_code_directive():
    sybil = Sybil([PythonCodeBlockParser()])
    check_path(sample_path('code.rst'), sybil, expected=1)


def test_sourcecode_directive():
    sybil = Sybil([PythonCodeBlockParser()])
    check_path(sample_path('sourcecode.rst'), sybil, expected=1)



import re

NEWLINE = re.compile("\n")


class LineNumberOffsets:

    def __init__(self, text: str) -> None:
        self.offsets = {
            line: match.start()+1 for (line, match) in enumerate(NEWLINE.finditer(text), start=1)
        }
        self.offsets[0] = 0

    def get(self, line: int, column: int) -> int:
        """
        Return the character offset of the  zero based line number and column offset.
        """
        return self.offsets[line] + column
